header { 
/** 
 * <copyright>
 * Copyright  2004 by Carnegie Mellon University, all rights reserved.
 *
 * Use of the Open Source AADL Tool Environment (OSATE) is subject to the terms of the license set forth
 * at http://www.eclipse.org/legal/cpl-v10.html.
 *
 * NO WARRANTY
 *
 * ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER PROPERTY OR RIGHTS GRANTED OR PROVIDED BY
 * CARNEGIE MELLON UNIVERSITY PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN "AS-IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING,
 * BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABILITY, INFORMATIONAL CONTENT,
 * NONINFRINGEMENT, OR ERROR-FREE OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, SPECIAL OR
 * CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE,
 * REGARDLESS OF WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. LICENSEE AGREES THAT IT WILL NOT
 * MAKE ANY WARRANTY ON BEHALF OF CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON CONCERNING THE
 * APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE DELIVERABLES UNDER THIS LICENSE.
 *
 * Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie Mellon University, its trustees, officers,
 * employees, and agents from all claims or demands made against them (and any related losses, expenses, or
 * attorney's fees) arising out of, or relating to Licensee's and/or its sub licensees' negligent use or willful
 * misuse of or negligent conduct or willful misconduct regarding the Software, facilities, or other rights or
 * assistance granted by Carnegie Mellon University under this License, including, but not limited to, any claims of
 * product liability, personal injury, death, damage to property, or violation of any laws or regulations.
 *
 * Carnegie Mellon University Software Engineering Institute authored documents are sponsored by the U.S. Department
 * of Defense under Contract F19628-00-C-0003. Carnegie Mellon University retains copyrights in all material produced
 * under this contract. The U.S. Government retains a non-exclusive, royalty-free license to publish or reproduce these
 * documents, or allow others to do so, for U.S. Government purposes only pursuant to the copyright license
 * under the contract clause at 252.227.7013.
 *
 * </copyright>
 *
 */
package edu.cmu.sei.aadl.parser; 
}
 
{	
	import edu.cmu.sei.aadl.model.core.*;
	import edu.cmu.sei.aadl.model.component.*;
	import edu.cmu.sei.aadl.model.feature.*;
	import edu.cmu.sei.aadl.model.connection.*;
	import edu.cmu.sei.aadl.model.instance.*;
	import edu.cmu.sei.aadl.model.property.*;
	import edu.cmu.sei.aadl.model.flow.*;
	import edu.cmu.sei.aadl.model.parsesupport.*; 
	import edu.cmu.sei.aadl.model.pluginsupport.*; 
	import edu.cmu.sei.aadl.annex.*; 
	import edu.cmu.sei.aadl.parser.provider.AadlParserPlugin;
	 
	import java.util.Vector;
	import java.util.Iterator;
	import java.util.Set;
	import java.util.HashSet;
	import org.eclipse.emf.ecore.util.EcoreUtil;

	import org.eclipse.emf.ecore.resource.Resource;
	import org.eclipse.core.resources.IResource; 

	import antlr.TokenStreamRecognitionException;
}

class AadlParser extends Parser;
options {
	buildAST=false;
	k=4;
//	codeGenBitsetTestThreshold=999;
	defaultErrorHandler=false;

}

{
	protected CoreFactory CoreF = CoreFactory.eINSTANCE;
	protected ComponentFactory CF = ComponentFactory.eINSTANCE;
	protected ConnectionFactory CnF = ConnectionFactory.eINSTANCE;
	protected FeatureFactory FF = FeatureFactory.eINSTANCE;
	protected FlowFactory FlF = FlowFactory.eINSTANCE;
	protected InstanceFactory IF = InstanceFactory.eINSTANCE;
	protected PropertyFactory PF = PropertyFactory.eINSTANCE;

	/**
	 * The error reporter to use.
	 */
	protected ParseErrorReporter errReporter = null;
	
	/**
	 * All external packages referenced in the parsed file.
	 */
	private Set referencedPackages = new HashSet();
	
	/**
	 * All packages defined iin the parsed file.
	 */
	private Set definedPackages = new HashSet();
	
	/**
	 * All external property sets referenced in the parsed file.
	 */
	private Set referencedPropertySets = new HashSet();

	/**
	 * All property sets defined in the parsed file.
	 */
	private Set definedPropertySets = new HashSet();

	/**
	 * Set the error reporter to use.
	 */
		
	public void setParseErrorReporter(final ParseErrorReporter reporter) {
		  errReporter = reporter;
	}
	
	/**
	 * Report parser error
	 * @param msg  error message
	 * @param e  antlr error exception 
	 */
	public  void reportError(RecognitionException e)  {
		  errReporter.error(e.getFilename(), e.getLine(), e.getMessage());
	}

	/**
	 * Report parser error
	 * @param msg  error message
	 * @param e  any exception 
	 */
	public  void reportError(Exception e) {
		AadlParserPlugin.INSTANCE.log(e);
		  errReporter.error(this.getFilename(), 0, e.getMessage());
	}
	/**
	 * Report parser error
	 * @param msg  error message
	 * @param e  antlr error exception 
	 */
	public  void reportError(ANTLRException e) {
			  errReporter.error(this.getFilename(), 0, e.getMessage());
			}

	/**
	 * the current resource being parsed
	 */
	protected SplitStreamFilter filter = null;
	
	/**
	 * sets the filter for pulling comment tokens.
	 * @param filter  filter with comment tokens
	 */
	public void setFilter(SplitStreamFilter filter){
		this.filter = filter;
	}
	
	/* Process comments attached to the beginning of a declaration */
	public void processComments(AObject o){
		attachComments(o, getComments());
	}

	/* Get comments at the start of a declaration */
	public Vector getComments() {
		if (filter == null ) return null;
		return filter.getSplitTokens();
	}

	/* Attach comments to an object. */
	public void attachComments(AObject o, Vector cmt) {
		if (cmt == null) return;
		for ( Iterator i = cmt.iterator(); i.hasNext();)
		{
			Token t = (Token) i.next();
			String c = t.getText();
			if (c.startsWith("--")){
				c = c.substring(2);
			} else if (c.startsWith("/*")){
				c = c.substring(2, c.length()-2);
			}
			int idx = c.indexOf("\n");
			while (idx >= 0){
				if (idx == 0){
					c = c.substring(1);
				} else {
				  String com = c.substring(0, idx-1);
				  o.addComment(com);
				  c = c.substring(idx+1);
				}
				idx = c.indexOf("\n");
			}
			if (c.length()>0){
			  o.addComment(c);
			}
		}
	}


	/* Process comments attached to the end of a declaration */
	public void processPostComments(AObject o){
		processComments(o);
	}
	
	private void setToDo(Set packagesToDo, Set propertySetsToDo) {
		if (packagesToDo != null) {
			for (Iterator i = referencedPackages.iterator(); i.hasNext();) {
				String name = (String) i.next();
				boolean foundIt = false;
				for (Iterator ii = definedPackages.iterator(); ii.hasNext();) {
					if (name.equalsIgnoreCase((String)ii.next())) {
						foundIt = true;
						break;
					}
				}
				if (!foundIt)
				    packagesToDo.add(name);
				
			}
		}
		if (propertySetsToDo != null) {
			for (Iterator i = referencedPropertySets.iterator(); i.hasNext();) {
				String name = (String) i.next();
				boolean foundIt = false;
				for (Iterator ii = definedPropertySets.iterator(); ii.hasNext();) {
					if (name.equalsIgnoreCase((String)ii.next())) {
						foundIt = true;
						break;
					}
				}
				if ( ! foundIt)
					propertySetsToDo.add(name);
			}
		}
	}
}

//-------------------------------------------------------------------
//-------------------------------------------------------------------
//  AADL Syntax per AADL_V0.994 -- Aug2004
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 4  COMPONENT TYPES AND PACKAGES
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 4.1 AADL Specifications
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

aadl_specification [Resource r, Set packagesToDo, Set propertySetsToDo] returns [AadlSpec aadl = CoreF.createAadlSpec()] 
		options {defaultErrorHandler=true;} 
	{
		if (r != null) {
			r.getContents().add(aadl);
		}
		aadl.setLocationReference(this.getFilename(),1);
	}
:
	(aadl_declaration[aadl])+
	{ processComments(aadl);}
	EOF
	{
	     
		setToDo(packagesToDo, propertySetsToDo);
	}
;

//-------------------------------------------------------------------
aadl_declaration [AadlSpec aadl ]
		options {defaultErrorHandler=true;} 
      { Classifier a=null;   PropertySet ps=null;AnnexLibrary ax = null;}:
      ( a = component_classifier 
      		{   if( aadl.findClassifier(a.getName()) != null) errReporter.warning(a.getLocationReference(),"Duplicate component classifier");
      			aadl.addClassifier(a);
      			processComments(a);
      		}
      |  a = port_group_type { if( aadl.findClassifier(a.getName()) != null) errReporter.warning(a.getLocationReference(),"Duplicate port group type");
      			aadl.addClassifier(a);
      			processComments(a);} 
      |  ax = annex_library { 
      			if( aadl.findAnnexLibrary(ax.getName()) != null) errReporter.warning(ax.getLocationReference(),"Duplicate annex library");
      			aadl.addAnnexLibrary(ax);
      			processComments(ax);}
      | package_spec[aadl] 
      | ps = property_set {  
      			if( aadl.findPropertySet(ps.getName()) != null) errReporter.error(ps.getLocationReference(),"Duplicate property set");
      			if( aadl.findPublicAadlPackage(ps.getName())!=null || aadl.findPrivateOnlyAadlPackage(ps.getName())!=null) errReporter.error(ps.getLocationReference(),"Aadl Package with same name exists");
      			aadl.addPropertySet(ps);
      			processComments(ps);}
      )
;

//-------------------------------------------------------------------
component_classifier returns [ComponentClassifier aadldecl=null ]:
      ( aadldecl = thread_type |   aadldecl = thread_implementation
      | aadldecl = thread_group_type |   aadldecl = thread_group_implementation
      | aadldecl = system_type |   aadldecl = system_implementation
      | aadldecl = data_type |   aadldecl = data_implementation
      | aadldecl = subprogram_type |   aadldecl = subprogram_implementation
      | aadldecl = process_type |   aadldecl = process_implementation
      | aadldecl = processor_type |   aadldecl = processor_implementation
      | aadldecl = memory_type |   aadldecl = memory_implementation
      | aadldecl = bus_type |   aadldecl = bus_implementation
      | aadldecl = device_type |   aadldecl = device_implementation
      )
	 
;
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 4.2 Packages
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

// Standalone package in a file
package_file [Resource r, Set packagesToDo, Set propertySetsToDo] returns [AadlSpec as=CoreF.createAadlSpec()] 
		options {defaultErrorHandler=true;} 
		{
			AadlPackage pkg = CoreF.createAadlPackage();
			String packname=null;
			String pname2=null;
			processComments(pkg);
		as.addAadlPackage(pkg);
		if (r != null) {
			r.getContents().add(as);
		}
		}
:
	fl:PACKAGE packname = package_name 
	{
		pkg.setName(packname);
		pkg.setLocationReference(this.getFilename(), fl.getLine());
	}
	(
		(public_part[pkg] (private_part[pkg])?)
	|
		private_part[pkg]
	)
	END pname2 = package_name sl:SEMI   
	{
		if (!(packname.equalsIgnoreCase(pname2))) {
			throw new SemanticException("package name '" + packname + "' does not match end identifier '" + pname2+"'", 
					getFilename(), sl.getLine(), sl.getColumn());
		}
		processComments(pkg);
		setToDo(packagesToDo, propertySetsToDo);
	}
	EOF
;

// Nested package inside an AADL specification
package_spec [AadlSpec aadl] 
		{AadlPackage aadllib = CoreF.createAadlPackage();
    	String packname=null;String pname2=null;
    	processComments(aadllib);}:
  fl:PACKAGE packname = package_name 
  			{ aadllib.setName(packname);
  			aadllib.setLocationReference(this.getFilename(),fl.getLine());
     			aadl.addAadlPackage(aadllib);}       
     ( (public_part[aadllib] ( private_part[aadllib] )?)
	 | private_part[aadllib] )
  END pname2 = package_name sl:SEMI   
  { 
     if( aadl.hasAadlPackage(aadllib) ) errReporter.error(aadllib.getLocationReference(),"Duplicate package");
     if( aadl.findPropertySet(aadllib.getName()) != null) errReporter.error(aadllib.getLocationReference(),"PropertySet with same name exists");
  	if (!( packname.equalsIgnoreCase(pname2) )) throw new SemanticException("package name '"+packname+"' does not match end identifier '"+pname2+"'", this.getFilename(),sl.getLine(), sl.getColumn());}
  { processComments(aadllib); }
  ;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

//-------------------------------------------------------------------
  
  public_part [AadlPackage aadllib]  
      { AadlPublic pub = CoreF.createAadlPublic(); aadllib.setAadlPublic(pub); processComments(pub);}:
      PUBLIC package_items[pub] ;
//-------------------------------------------------------------------
  
  private_part [AadlPackage aadllib]  
        { AadlPrivate priv = CoreF.createAadlPrivate(); aadllib.setAadlPrivate(priv); processComments(priv);}:
      PRIVATE package_items[priv] ;
//-------------------------------------------------------------------
 
 
package_items [AadlPackageSection section]
			{ Properties ps=null;} :
       ( package_item[section]
       )+
	   ( ps = propertyAssociations { section.setProperties(ps); })?
	   ;
	   
package_item [AadlPackageSection section]
		options {defaultErrorHandler=true;} 
         { Classifier cd = null; AnnexLibrary ax = null; }
     :  
             cd =  component_classifier { 
             	if( section.findClassifier(cd.getName()) != null) errReporter.warning(cd.getLocationReference(),"Duplicate component classifier");
             	section.addClassifier(cd);} 
         |
             cd = port_group_type { 
             	if( section.findClassifier(cd.getName()) != null) errReporter.warning(cd.getLocationReference(),"Duplicate port group type");
             	section.addClassifier(cd);}
         |
             ax = annex_library { 
             	if( section.findAnnexLibrary(ax.getName()) != null) errReporter.warning(ax.getLocationReference(),"Duplicate annex library");
      			section.addAnnexLibrary(ax);}
	;

package_name returns [String name = "";]:
  id:IDENT { name = id.getText();} ( DOUBLECOLON id2:IDENT { name = name + "::" + id2.getText();} )*
  {
  	definedPackages.add(name);
  }
  ;


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 4.3 Component Types
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

	
// note: we recognize none by the appropriate subclause object without content
none_stmt : NONE SEMI ; 

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ category specific type portion
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
thread_type returns [ThreadType tt=null]
		options {defaultErrorHandler=true;} 
   { ClassifierReference cr =null;
    AnnexSubclause ax=null; Properties ps=null; }:
	THREAD 	id:IDENT 	
	    { tt = CF.createThreadType(); processComments(tt);}
	    {tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	( featuresThread[tt] )?
	 ( flow_specs[tt] )?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT sl:SEMI
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

	
thread_group_type returns [ThreadGroupType tt=null]
options {defaultErrorHandler=true;}
   { AnnexSubclause ax=null; Properties ps=null; 
    ClassifierReference cr =null;}:
	THREAD GROUP 	id:IDENT 	
	    { tt = CF.createThreadGroupType(); processComments(tt);}
	    {	tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	(featuresThreadGroup[tt])?
	 ( flow_specs[tt] )?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT SEMI 
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

	
process_type returns [ProcessType tt=null]
options {defaultErrorHandler=true;}
   { AnnexSubclause ax=null; Properties ps=null; 
    ClassifierReference cr=null ;}:
	PROCESS 	id:IDENT 	
	    { tt = CF.createProcessType(); processComments(tt);}
	    {	tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS  cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	(featuresProcess[tt])?
	 ( flow_specs[tt] )?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT SEMI 
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }
	
system_type returns [SystemType tt=null]
options {defaultErrorHandler=true;}
   { AnnexSubclause ax=null; Properties ps=null; 
   ClassifierReference cr =null;}:
	SYSTEM 	id:IDENT 	
	    { tt = CF.createSystemType(); processComments(tt);}
	    {	tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	(featuresSystem[tt])?
	 ( flow_specs[tt] )?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT SEMI 
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }
	
subprogram_type returns [SubprogramType tt=null]
options {defaultErrorHandler=true;}
   { AnnexSubclause ax=null; Properties ps=null; 
   ClassifierReference cr =null;}:
	SUBPROGRAM 	id:IDENT 	
	    { tt = CF.createSubprogramType(); processComments(tt);}
	    {	tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	(featuresSubprogram[tt])?
	 ( flow_specs[tt] )?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT SEMI 
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

	
data_type returns [DataType tt=null]
options {defaultErrorHandler=true;}
   { AnnexSubclause ax=null; Properties ps=null; 
    ClassifierReference cr =null;}:
	DATA 	id:IDENT 	
	    { tt = CF.createDataType(); processComments(tt);}
	    {	tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	(featuresData[tt])?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT SEMI 
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

	
processor_type returns [ProcessorType tt=null]
options {defaultErrorHandler=true;}
   { AnnexSubclause ax=null; Properties ps=null; 
    ClassifierReference cr =null;}:
	PROCESSOR 	id:IDENT 	
	    { tt = CF.createProcessorType(); processComments(tt);}
	    {	tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	(featuresProcessor[tt])?
	 ( flow_specs[tt] )?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT SEMI 
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }
	
memory_type returns [MemoryType tt=null]
options {defaultErrorHandler=true;}
   { AnnexSubclause ax=null; Properties ps=null; 
    ClassifierReference cr =null;}:
	MEMORY 	id:IDENT 	
	    { tt = CF.createMemoryType(); processComments(tt);}
	    {	tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	(featuresMemory[tt])?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT SEMI 
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }
	
bus_type returns [BusType tt=null]
options {defaultErrorHandler=true;}
   { AnnexSubclause ax=null; Properties ps=null; 
    ClassifierReference cr =null;}:
	BUS 	id:IDENT 	
	    { tt = CF.createBusType(); processComments(tt);}
	    {	tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	(featuresBus[tt])?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT SEMI 
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }
	
device_type returns [DeviceType tt=null]
options {defaultErrorHandler=true;}
   { AnnexSubclause ax=null; Properties ps=null; 
    ClassifierReference cr =null;}:
	DEVICE 	id:IDENT 	
	    { tt = CF.createDeviceType(); processComments(tt);}
	    {	tt.setName(id.getText());tt.setLocationReference(this.getFilename(),id.getLine());}
	(EXTENDS cr = unique_type_name { tt.setExtendedClassifierReference(cr);})?
	(featuresDevice[tt])?
	 ( flow_specs[tt] )?
	( ps = propertyAssociations_no_modes { tt.setProperties(ps); } )?
	( ax = annex_subclause { tt.addAnnexSubclause(ax); } )?
    END eid:IDENT SEMI 
    { if (!id.getText().equalsIgnoreCase(eid.getText())) throw new SemanticException("End identifier "+eid.getText()+" does not match defining identifier "+id.getText(),this.getFilename(),eid.getLine(),eid.getColumn());}
    { processPostComments(tt); }
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }
		
//-------------------------------------------------------------------
//  unique type name must refer to a type
// unqiue implemenation name must refer to an implementation
// classifier reference may be incomplete.
// ClassifierReference is an object that has three fields:
// package name, type name, impl name. Any of them can be null
//---------------------------------------------------------------

unique_type_name returns [ClassifierReference cr = new ClassifierReference()]
{String pn = "";}:
	 ( pid: IDENT DOUBLECOLON 
	 		{ pn = pn + (pn.length() == 0 ? "":"::")+pid.getText();}
	 )* 
	 {
	 	if ( pn.length() > 0 ) {
	 		cr.setPackageName(pn);
	 		referencedPackages.add(pn);
	 	}
	 }
	 tid:IDENT { cr.setClassifierName(tid.getText()); cr.setFilename(this.getFilename());cr.setLine(tid.getLine());}
;

unique_impl_name returns [ClassifierReference cr = new ClassifierReference()]
{String pn = "";}:
	 ( pid: IDENT DOUBLECOLON 
	 		{ pn = pn +(pn.length() == 0 ? "":"::")+pid.getText();}
	 )* 
	 {
	 	if ( pn.length() > 0 ) {
	 		cr.setPackageName(pn);
	 		referencedPackages.add(pn);
	 	}
	 }
	 tid:IDENT { cr.setFilename(this.getFilename());cr.setLine(tid.getLine());}
	DOT iid:IDENT { cr.setClassifierName(tid.getText()+"."+iid.getText());}
    ;

classifier_reference returns [ClassifierReference cr = new ClassifierReference()] 
{String pn = "";}:
	 ( pid: IDENT DOUBLECOLON 
	 		{ pn = pn +(pn.length() == 0 ? "":"::")+pid.getText();}
	 )* 
	 {
	 	if ( pn.length() > 0 ) {
	 		cr.setPackageName(pn);
	 		referencedPackages.add(pn);
	 	}
	 }
	 tid:IDENT {cr.setClassifierName(tid.getText()); cr.setFilename(this.getFilename());cr.setLine(tid.getLine());}
	( DOT iid:IDENT 
			{ cr.setClassifierName(tid.getText()+"."+iid.getText());}
	)?
    ;

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 4.4 Component Implementations
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ category specific implementation portion
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

common_component_impl [ComponentImpl ci]
{Properties ps =null;AnnexSubclause ax =null;}
:
		(mode_subclause[ci])?
	   ( ps = containedPropertyAssociations { ci.setProperties(ps); processComments(ps);} )?
		( ax = annex_subclause { ci.addAnnexSubclause(ax); processComments(ax);} )?

;


thread_implementation returns [ThreadImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null; CallSequences calls = null;}:
		THREAD IMPL { ci = CF.createThreadImpl();processComments(ci);}
	 typeid:IDENT DOT 
	defid:IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	   (EXTENDS cr = unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		(threadSubcomponents[ci] )?
		(calls = callsSubclause { ci.setCallSequences(calls);})?
		(connectionsSubclause[ci])?
		(flow_impls[ci])?
		common_component_impl[ci]

		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("Defining identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }


thread_group_implementation returns [ThreadGroupImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null;}:
		THREAD GROUP IMPL { ci = CF.createThreadGroupImpl();processComments(ci);}
	 typeid:IDENT DOT 
	defid:IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	(EXTENDS cr = unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		(threadgroupSubcomponents[ci] )?
		(connectionsSubclause[ci])?
		(flow_impls[ci])?
		common_component_impl[ci]

		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }


process_implementation returns [ProcessImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null;}:
		PROCESS IMPL { ci = CF.createProcessImpl();processComments(ci);}
	 typeid: IDENT DOT 
	 defid: IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	   (EXTENDS cr= unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		(processSubcomponents[ci] )?
		(connectionsSubclause[ci])?
		(flow_impls[ci])?
		common_component_impl[ci]

		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }


system_implementation returns [SystemImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null;}:
		SYSTEM IMPL { ci = CF.createSystemImpl();processComments(ci);}
	 typeid: IDENT DOT 
	defid: IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	  (EXTENDS cr= unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		(systemSubcomponents[ci] )?
		(connectionsSubclause[ci])?
		(flow_impls[ci])?
		common_component_impl[ci]


		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }


data_implementation returns [DataImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null;}:
		DATA IMPL { ci = CF.createDataImpl();processComments(ci);}
	 typeid: IDENT DOT 
	defid: IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	  (EXTENDS cr= unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		(dataSubcomponents[ci] )?
        (dataConnectionsSubclause[ci] )?		
        common_component_impl[ci]

		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }


subprogram_implementation returns [SubprogramImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null; CallSequences calls=null;}:
		SUBPROGRAM IMPL { ci = CF.createSubprogramImpl();processComments(ci);}
	 typeid: IDENT DOT 
	defid: IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	(EXTENDS cr= unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		( calls = callsSubclause {ci.setCallSequences(calls);} )?
		(connectionsSubclause[ci])?
		(flow_impls[ci])?
		common_component_impl[ci]

		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }


processor_implementation returns [ProcessorImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null;}:
		PROCESSOR IMPL { ci = CF.createProcessorImpl();processComments(ci);}
	 typeid: IDENT DOT 
	defid: IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	(EXTENDS cr= unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		(processorSubcomponents[ci])? 
        (busConnectionsSubclause[ci] )?		
		(flow_impls[ci])?
		common_component_impl[ci]

		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

memory_implementation returns [MemoryImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null;}:
		MEMORY IMPL { ci = CF.createMemoryImpl();processComments(ci);}
	 typeid: IDENT DOT 
	defid: IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	(EXTENDS cr= unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		(memorySubcomponents[ci] )?
        (busConnectionsSubclause[ci] )?		
		common_component_impl[ci]

		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

bus_implementation returns [BusImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null;}:
		BUS IMPL { ci = CF.createBusImpl();processComments(ci);}
	 typeid: IDENT DOT 
	defid: IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	(EXTENDS cr= unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		common_component_impl[ci]

		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

device_implementation returns [DeviceImpl ci=null ]
options {defaultErrorHandler=true;}
{ClassifierReference cr=null;}:
		DEVICE IMPL { ci = CF.createDeviceImpl();processComments(ci);}
	 typeid: IDENT DOT 
	defid: IDENT { ci.setName(typeid.getText()+"."+defid.getText());ci.setLocationReference(this.getFilename(),defid.getLine());}
	(EXTENDS cr = unique_impl_name { ci.setExtendedClassifierReference(cr);})?
	    (refinestypeSubclause[ci])?
		(flow_impls[ci])?
		common_component_impl[ci]

		END id:IDENT DOT id2:IDENT SEMI
           { if (!((id.getText().equalsIgnoreCase(typeid.getText()) && id2.getText().equalsIgnoreCase(defid.getText()) ))) throw new SemanticException("identifier "+typeid.getText()+"."+defid.getText()+" does not match end identifier "+id.getText()+"."+id2.getText(),this.getFilename(),id.getLine(),id.getColumn());}
            { processPostComments(ci); }
		;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ category specific provides clauses
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

featuresData [DataType ct]  
{DataFeatures pr=null;}:		 
     FEATURES {  pr = FF.createDataFeatures(); processComments(pr);}
                   { ct.setFeatures(pr); }
	  ( ( 
	  		dataFeature[pr]
	    )+ 
	      // an empty provides object represents none
		|  none_stmt )
	;
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

dataFeature[DataFeatures pr]
	        {boolean refined = false; Feature fs =null;}:
         id: IDENT 
         COLON 
         ( REFINED TO { refined = true; } )?
	      
	    (  fs=data_subprogram 
        | ( PROVIDES fs= dataAccess[AccessDirection.PROVIDED_LITERAL]
	      )
	    ) 
                { fs.setName(id.getText()); fs.setRefined(refined); pr.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
				{ if (refined){RefinedReference fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}}
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


featuresSubprogram [SubprogramType ct]  
//options {defaultErrorHandler=true;}
{SubprogramFeatures pr=null;}:		 
     FEATURES {  pr = FF.createSubprogramFeatures(); processComments(pr);}
                   { ct.setFeatures(pr); }
	    ( ( subprogramFeature[pr]
	      )+ 
	    // an empty provides object represents none
		|  none_stmt  )
	;
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

subprogramFeature[SubprogramFeatures pr]
	        {boolean refined = false; Feature fs =null;}:
         id: IDENT 
         COLON 
         ( REFINED TO { refined = true; } )?
	      
	      ( fs= subprogramPort 
	        | fs = port_group_spec 
	      |
	        ( REQUIRES fs= dataAccess[AccessDirection.REQUIRED_LITERAL] 
	        )
	      | fs= parameter 
	      )
                { fs.setName(id.getText()); fs.setRefined(refined); pr.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                { if (refined){
                	RefinedReference fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}
                }
;		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

featuresThread [ThreadType ct]  
//options {defaultErrorHandler=true;}
{ ThreadFeatures pr = FF.createThreadFeatures(); processComments(pr);
         ct.setFeatures(pr); }:		 
    FEATURES 
	    ( ( threadFeature[pr]
	      )+ 
	    // an empty provides object represents none
		|  none_stmt  )
    ;
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

threadFeature[ThreadFeatures pr]
	        {boolean refined = false; Feature fs =null;}:
         id:IDENT 
         COLON 
         ( REFINED TO { refined = true; } )?
	      
	      (  fs= port_spec 
	        | fs = port_group_spec 
	        | fs = server_subprogram 
	        | (PROVIDES fs= dataAccess[AccessDirection.PROVIDED_LITERAL] )
	        | (REQUIRES fs= dataAccess[AccessDirection.REQUIRED_LITERAL] )
	        )
                { fs.setName(id.getText()); fs.setRefined(refined); pr.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                { if (refined){RefinedReference fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}}
;		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


featuresThreadGroup [ThreadGroupType ct]  
//options {defaultErrorHandler=true;}
{ ThreadGroupFeatures pr = FF.createThreadGroupFeatures(); processComments(pr);
                    ct.setFeatures(pr); }:		 
    FEATURES 
	    ( ( threadGroupFeature[pr]
	    )+ 
	    // an empty provides object represents none
		|  none_stmt  );
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

threadGroupFeature[ThreadGroupFeatures pr]
	        {boolean refined = false; RefinedReference fr=null;Feature fs =null;}:
         id: IDENT 
         COLON 
         ( REFINED TO { refined = true; } )?
	      
	      (  fs= port_spec 
	      | fs= port_group_spec 
	      | fs= server_subprogram 
	      | (PROVIDES fs=dataAccess[AccessDirection.PROVIDED_LITERAL] )
	      | (REQUIRES fs= dataAccess[AccessDirection.REQUIRED_LITERAL] )
	        )
                { fs.setName(id.getText()); fs.setRefined(refined); pr.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}}
;		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

featuresProcess [ProcessType ct]  
//options {defaultErrorHandler=true;}
{ ProcessFeatures pr = FF.createProcessFeatures(); processComments(pr);
                    ct.setFeatures(pr); }:		 
    FEATURES 
	    ( ( processFeature[pr]
	    )+ 
	    // an empty provides object represents none
		|  none_stmt );
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

processFeature[ProcessFeatures pr]
	        {boolean refined = false; RefinedReference fr=null;Feature fs =null;}:
         id: IDENT 
         COLON 
         ( REFINED TO { refined = true; } )?
	      
	      (  fs= port_spec 
	      | fs= port_group_spec 
	      | fs= server_subprogram 
	      | (PROVIDES fs= dataAccess[AccessDirection.PROVIDED_LITERAL] )
	      | (REQUIRES fs= dataAccess[AccessDirection.REQUIRED_LITERAL] )
	      )
                { fs.setName(id.getText()); fs.setRefined(refined); pr.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}}
;		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


featuresSystem [SystemType ct]  
//options {defaultErrorHandler=true;}
{ SystemFeatures pr = FF.createSystemFeatures(); processComments(pr);
                   ct.setFeatures(pr); }:		 
     FEATURES 
	    ((  systemFeature[pr]
	     )+ 
	    // an empty provides object represents none
		|  none_stmt  );
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

systemFeature[SystemFeatures pr]
	        {boolean refined = false; RefinedReference fr=null;Feature fs =null;}:
           id: IDENT 
           COLON 
           ( REFINED TO { refined = true; } )?
	      
	       (  fs= port_spec 
	         | fs= port_group_spec 
	         | fs= server_subprogram 
	         | ( PROVIDES fs=  dataAccess[AccessDirection.PROVIDED_LITERAL] )
	         | (PROVIDES fs= busAccess[AccessDirection.PROVIDED_LITERAL] )
	         | (REQUIRES fs= dataAccess[AccessDirection.REQUIRED_LITERAL] )
	         | (REQUIRES fs= busAccess[AccessDirection.REQUIRED_LITERAL] )
	        )
                { fs.setName(id.getText()); fs.setRefined(refined); pr.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}}
;		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


featuresProcessor[ProcessorType ct]  
//options {defaultErrorHandler=true;}
{ ProcessorFeatures pr = FF.createProcessorFeatures(); processComments(pr);
                    ct.setFeatures(pr); }:		 
     FEATURES 
	  (  (  
	  		processorFeature[pr]
	    )+ 
	    // an empty provides object represents none
		|  none_stmt );
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

processorFeature[ProcessorFeatures pr]
	        {boolean refined = false; RefinedReference fr=null;Feature fs =null;}:
         id: IDENT 
         COLON 
         ( REFINED TO { refined = true; } )?
	      
	      (  fs= port_spec 
	      | fs= port_group_spec 
	      | fs= server_subprogram 
	      | (REQUIRES fs= busAccess[AccessDirection.REQUIRED_LITERAL] )
	      )
                { fs.setName(id.getText()); fs.setRefined(refined); pr.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}}
;		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

featuresDevice [DeviceType ct] 
//options {defaultErrorHandler=true;}
{ DeviceFeatures pr = FF.createDeviceFeatures(); processComments(pr);
                    ct.setFeatures(pr); }:
     FEATURES 
	  (  (  
	  		deviceFeature[pr]
	    )+ 
	    // an empty provides object represents none
		|  none_stmt 
	);
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

deviceFeature[DeviceFeatures pr]
	        {boolean refined = false; Feature fs =null;}:
           id: IDENT 
           COLON 
           ( REFINED TO { refined = true; } )?
	      
	       ( fs=  port_spec 
	       | fs= port_group_spec 
	       | fs= server_subprogram 
	       |( REQUIRES fs= busAccess[AccessDirection.REQUIRED_LITERAL] )
	       |( PROVIDES fs= busAccess[AccessDirection.PROVIDED_LITERAL] )
	       )
                { fs.setName(id.getText()); fs.setRefined(refined); pr.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                { if (refined){RefinedReference fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}}
;		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }
	

featuresMemory [MemoryType ct]  
//options {defaultErrorHandler=true;}
{ MemoryFeatures pr = ct.getFeatures();}:	
     FEATURES { if (pr == null) {pr = FF.createMemoryFeatures(); ct.setFeatures(pr);} processComments(pr);}
	    (  (
	    memoryFeature[pr]
	      )+ 
	      // an empty provides object represents none
		  |  none_stmt  
		);
		exception 
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

memoryFeature[MemoryFeatures pr]
		        {boolean refined = false; Feature fs =null;}:
           id: IDENT 
           COLON 
           ( REFINED TO { refined = true; } )?
	       (
	         REQUIRES fs= busAccess[AccessDirection.REQUIRED_LITERAL] 
                { fs.setName(id.getText()); fs.setRefined(refined); pr.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                { if (refined){RefinedReference fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}}
	       )
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


featuresBus [BusType ct]  
//options {defaultErrorHandler=true;}
            { BusFeatures pr = ct.getFeatures();}:
     FEATURES { if (pr == null) {pr = FF.createBusFeatures(); ct.setFeatures(pr);}processComments(pr); }
     ( (
	     busFeature[pr]
	    )+ 
	    // an empty provides object represents none
		|  none_stmt 
	) ;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

busFeature[BusFeatures bf]
	        {boolean refined = false; Feature fs =null;}:
         id: IDENT 
         COLON 
         ( REFINED TO { refined = true; } )?
	      (
	        REQUIRES fs= busAccess[AccessDirection.REQUIRED_LITERAL] 
	      )
                { fs.setName(id.getText()); fs.setRefined(refined); bf.addFeature(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                { if (refined){RefinedReference fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}}
   ;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ refines type subclause
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

refinestypeSubclause [ComponentImpl ci]
{ RefinesType rt = CoreF.createRefinesType(); 
	ci.setRefinesType(rt);processComments(rt);
	RefinedReference fr=null;Feature fs =null;}: 
	REFINES TYPE 
		( ( 
             id: IDENT 
             COLON 
             REFINED TO 
			 ( fs =  port_spec 
			 | fs = port_group_spec
			 | fs = subcomponent_access
			 | fs = parameter 
			 | fs = server_subprogram
			 | fs = data_subprogram
			 )
                { fs.setName(id.getText()); fs.setRefined(true); rt.addFeatureRefinement(fs); }
				{ fs.setLocationReference(this.getFilename(),id.getLine());  }
                {  fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	fs.setRefinedReference(fr);}
		   )+ 
		 | ( none_stmt ) 
		 )
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 4.5 Subcomponents
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ category specific subcompoennts subclause to reflect acceptable categories
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
threadSubcomponents [ThreadImpl ci]
{ ThreadSubcomponents ts = CF.createThreadSubcomponents(); processComments(ts);}:		 
     SUBCOMPONENTS 
                     {  ci.setSubcomponents(ts); }
			( ( threadSubcomponent[ts] )+ 
			| ( none_stmt ) )
	;
			
processorSubcomponents[ProcessorImpl ci]
{ ProcessorSubcomponents ts = CF.createProcessorSubcomponents(); processComments(ts);}: 
    SUBCOMPONENTS 
                     {  ci.setSubcomponents(ts); }
			( ( processorSubcomponent[ts] )+ 
			| ( none_stmt ) )
	;


memorySubcomponents[MemoryImpl ci]
{ MemorySubcomponents ts = CF.createMemorySubcomponents(); processComments(ts);}: 
     SUBCOMPONENTS 
                     {  ci.setSubcomponents(ts); }
			( ( memorySubcomponent[ts] )+ 
			| ( none_stmt ) )
	;


threadgroupSubcomponents[ThreadGroupImpl ci]
{ ThreadGroupSubcomponents ts = CF.createThreadGroupSubcomponents(); processComments(ts);}: 
     SUBCOMPONENTS 
                     {  ci.setSubcomponents(ts); }
			( ( threadgroupSubcomponent[ts] )+ 
			| ( none_stmt ) )
	;


processSubcomponents[ProcessImpl ci]
{ ProcessSubcomponents ts = CF.createProcessSubcomponents(); processComments(ts);}: 
     SUBCOMPONENTS 
                     {  ci.setSubcomponents(ts); }
			( ( processSubcomponent[ts] )+ 
			| ( none_stmt ) )
	;


systemSubcomponents[SystemImpl ci]
{ SystemSubcomponents ts = CF.createSystemSubcomponents(); processComments(ts);}: 
     SUBCOMPONENTS 
                     {  ci.setSubcomponents(ts); }
			( ( systemSubcomponent[ts] )+ 
			| ( none_stmt ) )
	;


dataSubcomponents[DataImpl ci]
{ DataSubcomponents ts = CF.createDataSubcomponents(); processComments(ts);}: 
     SUBCOMPONENTS 
                     {  ci.setSubcomponents(ts); }
			( ( dataSubcomponent[ts] )+ 
			| ( none_stmt ) )
	;
	


		
			
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ category specific subcomponent declaration to reflect acceptable categories
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

threadSubcomponent [ThreadSubcomponents ts]
{DataSubcomponent ds=null;
RefinedReference fr =null;}:
    id: IDENT 
    COLON 
        {ds = CF.createDataSubcomponent(); processComments(ds);}
    ( REFINED TO 
    	{  ds.setRefined(true);} 
        {  fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
        	fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
        	ds.setRefinedReference(fr);}
    )?
    DATA 
    subcomponent_decl[ds] 
            {  ds.setName(id.getText()); ts.addDataSubcomponent(ds); }
				{ ds.setLocationReference(this.getFilename(),id.getLine());  }
    ;

threadgroupSubcomponent [ThreadGroupSubcomponents ts]
        {Subcomponent sc = null; RefinedReference fr = null;
        	boolean refined = false;}:
    id: IDENT COLON 
    ( REFINED TO {  refined = true;} 
    )?
    ( DATA {sc = CF.createDataSubcomponent();processComments(sc);}
     |
      THREAD GROUP {sc = CF.createThreadGroupSubcomponent();processComments(sc);}
     |
      THREAD {sc = CF.createThreadSubcomponent();processComments(sc);}
     )
         {sc.setRefined(refined);}
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	sc.setRefinedReference(fr);}}
    subcomponent_decl[sc] 
            {  sc.setName(id.getText()); ts.addSubcomponent(sc); }
				{ sc.setLocationReference(this.getFilename(),id.getLine());  }
    ;

dataSubcomponent [DataSubcomponents ts]
        { boolean refined = false;
        	RefinedReference fr=null;
        	DataSubcomponent ds=null;}:
    id: IDENT 
    COLON 
    ( REFINED TO {  refined = true;} )?
    DATA { ds = CF.createDataSubcomponent();processComments(ds);}
         {ds.setRefined(refined);}
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	ds.setRefinedReference(fr);}}
    subcomponent_decl[ds] 
                {  ds.setName(id.getText()); ts.addDataSubcomponent(ds); }
				{ ds.setLocationReference(this.getFilename(),id.getLine());  }
    ;


processSubcomponent [ProcessSubcomponents ts]
        {Subcomponent sc = null;boolean refined = false;
        RefinedReference fr=null;}:
    id: IDENT 
    COLON 
    ( REFINED TO {  refined = true;} )?
    ( DATA {sc = CF.createDataSubcomponent(); processComments(sc);}
     |
      THREAD GROUP {sc = CF.createThreadGroupSubcomponent();processComments(sc);}
     |
      THREAD {sc = CF.createThreadSubcomponent();processComments(sc);}
      )
         {sc.setRefined(refined);}
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	sc.setRefinedReference(fr);}}
    subcomponent_decl[sc] 
                {  sc.setName(id.getText()); ts.addSubcomponent(sc); }
				{ sc.setLocationReference(this.getFilename(),id.getLine());  }
    ;


processorSubcomponent [ProcessorSubcomponents ts]
	{ MemorySubcomponent ds=null;
		boolean refined = false;RefinedReference fr=null;}:
    id: IDENT 
    COLON 
    ( REFINED TO {  refined = true;} )?
    MEMORY { ds = CF.createMemorySubcomponent();processComments(ds);}
         {ds.setRefined(refined);}
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	ds.setRefinedReference(fr);}}
    subcomponent_decl[ds] 
                {  ds.setName(id.getText()); ts.addMemorySubcomponent(ds); }
				{ ds.setLocationReference(this.getFilename(),id.getLine());  }
    ;

memorySubcomponent [MemorySubcomponents ts]
	{ MemorySubcomponent ds=null;
		boolean refined = false;RefinedReference fr=null;}:
    id: IDENT 
    COLON 
    ( REFINED TO {  refined = true;} )?
    MEMORY { ds = CF.createMemorySubcomponent();processComments(ds);}
         {ds.setRefined(refined);}
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	ds.setRefinedReference(fr);}}
    subcomponent_decl[ds] 
                {  ds.setName(id.getText()); ts.addMemorySubcomponent(ds); }
				{ ds.setLocationReference(this.getFilename(),id.getLine());  }
    ;

systemSubcomponent [SystemSubcomponents ts]
	{ boolean refined = false;RefinedReference fr=null;
        Subcomponent sc = null;}:
    id: IDENT 
    COLON 
    ( REFINED TO {  refined = true;} )?
    (
    DATA {sc = CF.createDataSubcomponent(); processComments(sc);}
    |
    MEMORY {sc = CF.createMemorySubcomponent(); processComments(sc);}
    |
    PROCESSOR {sc = CF.createProcessorSubcomponent();processComments(sc);}
    |
    BUS {sc = CF.createBusSubcomponent();processComments(sc);}
    |
    DEVICE {sc = CF.createDeviceSubcomponent();processComments(sc);}
    |
    PROCESS { sc = CF.createProcessSubcomponent();processComments(sc);}
    |
    SYSTEM { sc = CF.createSystemSubcomponent();processComments(sc);}
    )
         {sc.setRefined(refined);}
                { if (refined){ fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
           			fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
                	sc.setRefinedReference(fr);}}
    subcomponent_decl[sc] 
                {  sc.setName(id.getText()); ts.addSubcomponent(sc); }
				{ sc.setLocationReference(this.getFilename(),id.getLine());  }
    ;

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ common part of subcomponent declaration
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

subcomponent_decl[Subcomponent sc] 
		{ClassifierReference cr;Properties ps;}:
	( cr = classifier_reference  
			{ sc.setClassifierReference(cr);}
	)?
	(ps = containedCurlyPropertyAssociations {sc.setProperties(ps);})?
	 (in_modes[sc])? SEMI ;




//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 4.6 Annex Subclasses
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

// **** set up gammar to let everything but **} through

annex_library returns [AnnexLibrary al = null]
options {defaultErrorHandler=true;}
	{
		Vector cmt = null;
	}:
    ANNEX id:IDENT
		{
			cmt = getComments();
		}
	at : ANNEX_TEXT 
		{ 
			String annexName = id.getText();
			String annexText = at.getText();
			if (annexText.length() > 6) {
		        annexText = annexText.substring(3, annexText.length() - 3);
			}
			AnnexParserRegistry registry = (AnnexParserRegistry) AnnexRegistry.getRegistry(AnnexRegistry.ANNEX_PARSER_EXT_ID);
			AnnexParser ap = registry.getAnnexParser(annexName);
			al = ap.parseAnnexLibrary(id.getText(), annexText, getFilename(), at.getLine(), at.getColumn() + 3, errReporter);
			al.setName(id.getText());
			al.setLocationReference(this.getFilename(), id.getLine());
			attachComments(al, cmt);
	   }
	SEMI;

annex_subclause returns [AnnexSubclause as = null]
options {defaultErrorHandler=true;}
	{
		Vector cmt = null;
	}: 
    ANNEX id:IDENT
		{
			cmt = getComments();
		}
	at : ANNEX_TEXT
		{ 
			String annexName = id.getText();
			String annexText = at.getText();
			if (annexText.length() > 6 ){
		        annexText = annexText.substring(3, annexText.length() - 3);
			}
			AnnexParserRegistry registry = (AnnexParserRegistry) AnnexRegistry.getRegistry(AnnexRegistry.ANNEX_PARSER_EXT_ID);
			AnnexParser ap = registry.getAnnexParser(annexName);
			as = ap.parseAnnexSubclause(id.getText(), annexText, getFilename(), at.getLine(), at.getColumn() + 3, errReporter);
			as.setName(id.getText());
			as.setLocationReference(this.getFilename(), id.getLine());
			attachComments(as, cmt);
		}
	SEMI ;


 
 


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 5 PROPERTIES
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 5.1 Property Sets
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

// We must support multiple property type, name, and constant definitions in one declaration
// This is done through the list of defining identifiers
// A pain in the neck to process. We must first construct the whole thing, then dupliate it.
// The addition of the items into the propertyset is subclass specific, thus, ps and the list of identifeir strings are passed in as paramters

// Standalone property set in a separate file
propertyset_file [Resource r, Set packagesToDo, Set propertySetsToDo] returns [AadlSpec as=CoreF.createAadlSpec()] 
		options {defaultErrorHandler=true;}
		{
		PropertySet ps=PF.createPropertySet();
			String ps_name = null;
			String ps_end_name = null;
		as.addPropertySet(ps);
		if (r != null) {
			r.getContents().add(as);
		}
		} 
:
	PROPERTY SET ps_id:IDENT IS
	{
		ps_name = ps_id.getText();
		ps.setName(ps_name);
		definedPropertySets.add(ps_name);
		ps.setLocationReference(this.getFilename(), ps_id.getLine());
		processComments(ps);
	}
	(   
		id: IDENT COLON
		{
			if (ps.findPropertySetElement(id.getText()) != null ) {
				errReporter.warning(new LocationReference(this.getFilename(), id.getLine()), "Duplicate property set entry");
			}
		}
		(
			property_constant[ps, id] 
			| property_type_declaration[ps, id] 
			| property_name_declaration[ps, id]
		)
	)+
	END ps_id2:IDENT SEMI
	{
		ps_end_name = ps_id2.getText();
		if (!(ps_end_name.equalsIgnoreCase(ps_name))) {
			throw new SemanticException("Property set name '" + ps_name + "' does not match end identifier '" + ps_end_name+"'", 
					getFilename(), ps_id2.getLine(), ps_id2.getColumn());
		}
		processComments(ps);
	}
	EOF
	{
//		System.out.println("Referenced Packages: ");
//		for (Iterator i = referencedPackages.iterator(); i.hasNext();) {
//			System.out.println("\t" + (String)i.next());
//		}
//		System.out.println("\nDefined Packages: ");
//		for (Iterator i = definedPackages.iterator(); i.hasNext();) {
//			System.out.println("\t" + (String)i.next());
//		}
//		System.out.println("\nReferenced Property Sets: ");
//		for (Iterator i = referencedPropertySets.iterator(); i.hasNext();) {
//			System.out.println("\t" + (String)i.next());
//		}
//		System.out.println("\nDefined Property Sets: ");
//		for (Iterator i = definedPropertySets.iterator(); i.hasNext();) {
//			System.out.println("\t" + (String)i.next());
//		}
		setToDo(packagesToDo, propertySetsToDo);
		
	}
;


// Nested property set inside an Aadl specification
property_set returns [PropertySet ps=PF.createPropertySet()]
		options {defaultErrorHandler=true;}
		{
			String ps_name = null;
			String ps_end_name = null;
		}
:
	PROPERTY SET ps_id: IDENT IS
	{
		ps_name = ps_id.getText();
		ps.setName(ps_name);
		definedPropertySets.add(ps_name);
		ps.setLocationReference(this.getFilename(), ps_id.getLine());
		processComments(ps);
	}
	(   
		id: IDENT COLON
		{
			if (ps.findPropertySetElement(id.getText()) != null ) {
				errReporter.warning(new LocationReference(this.getFilename(), id.getLine()), "Duplicate property set entry");
			}
		}
		(
			property_constant[ps, id] 
			| property_type_declaration[ps, id] 
			| property_name_declaration[ps, id]
		)
	)+
	END ps_id2:IDENT SEMI
	{
		ps_end_name = ps_id2.getText();
		if (!(ps_end_name.equalsIgnoreCase(ps_name))) {
			throw new SemanticException("Property set name '" + ps_name + "' does not match end identifier '" + ps_end_name+"'", 
					getFilename(), ps_id2.getLine(), ps_id2.getColumn());
		}
		processComments(ps);
	};
	exception
		catch [RecognitionException ex] {
			reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 5.1.1 Property Types
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

// Property type can be named, i.e., defined with a list of identifiers.
// this means they must be replicated.
// They can also be declared as unnamed types as part of property name declarations.
// In that case they are a single instance with no name. Thus, two rules for every property type.

// ParsedPropertyReference object is used to hold the parsed idendieirs. 
// This is used for references to other properties, from property associaitons to property names,
// references to units literals, and enumeration literals. The objects are processed by the name resolver.

property_type_declaration [PropertySet ps,Token il] 
options {defaultErrorHandler=true;}:
    TYPE 
    property_type[ps,il] 
    SEMI
//-------------------------------------------------------------------
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


property_type [PropertySet ps, Token il]:
    boolean_type[ps,il]
    | string_type[ps,il]
	| enumeration_type[ps,il] | units_type[ps,il]
	| real_type[ps,il] | integer_type[ps,il] 
	| range_type[ps,il]
	| classifier_type[ps,il]
    | component_property_type[ps,il]
//-------------------------------------------------------------------
;

unnamed_property_type returns [PropertyType pt=null]:
    pt = unnamed_boolean_type
    | pt = unnamed_string_type
	| pt = unnamed_enumeration_type | pt = unnamed_units_type
	| pt = unnamed_real_type | pt = unnamed_integer_type 
	| pt = unnamed_range_type
	| pt = unnamed_classifier_type
    | pt = unnamed_component_property_type
//-------------------------------------------------------------------
;

unnamed_boolean_type  returns [Aadlboolean ab=PF.createAadlboolean()]:
    BOOLEAN  { processComments(ab);}
   ;

boolean_type [PropertySet ps, Token il]
{Aadlboolean pt=null; LocationReference lr=null;}:
    BOOLEAN 
      { pt = PF.createAadlboolean();pt.setName(il.getText());ps.addPropertyType(pt); processComments(pt);}
         { lr= new LocationReference(this.getFilename(),il.getLine()); pt.setLocationReference(lr);}
   ;

string_type  [PropertySet ps, Token il]
{Aadlstring pt=null; LocationReference lr=null;}:
    STRING 
      {  pt = PF.createAadlstring();pt.setName(il.getText());ps.addPropertyType(pt);  processComments(pt);}
         {  lr = new LocationReference(this.getFilename(),il.getLine()); pt.setLocationReference(lr);}
   ;
   
unnamed_string_type  returns [Aadlstring ab=PF.createAadlstring()]:
    STRING  {processComments(ab);}
   ;
   
enumeration_type [PropertySet ps, Token il]
		{ EnumType et; LocationReference lr=null;}:
	et = unnamed_enumeration_type
	  {  et.setName(il.getText()); ps.addPropertyType(et);processComments(et);}
         {  lr = new LocationReference(this.getFilename(),il.getLine()); et.setLocationReference(lr);}
;

unnamed_enumeration_type returns [EnumType et = PF.createEnumType()]
{ EnumLiteral el=null; }:
    ENUMERATION
	LPAREN lit: IDENT  
	    { el = PF.createEnumLiteral();
	      el.setName(lit.getText());
	      el.setLocationReference(this.getFilename(),lit.getLine());
	      et.addEnumLiteral(el);
	    }
	( COMMA morelit:IDENT 
	    { el = PF.createEnumLiteral();
	      el.setName(morelit.getText());
	      el.setLocationReference(this.getFilename(), morelit.getLine());
	      et.addEnumLiteral(el); }
	)* RPAREN 
;
	  
//-------------------------------------------------------------------

units_type [PropertySet ps, Token il]
        { UnitsType ut = PF.createUnitsType();LocationReference lr=null;}:
    UNITS units_list[ut]
	  { ut.setName(il.getText()); ps.addPropertyType(ut);processComments(ut);}
         {  lr = new LocationReference(this.getFilename(),il.getLine()); ut.setLocationReference(lr);}
;

unnamed_units_type returns [UnitsType ut=PF.createUnitsType()]:
	UNITS units_list[ut]
	;

units_list [UnitsType ut]
{UnitLiteral el=null; NumberValue nv = null;}:
    LPAREN lit:IDENT
	    { el = PF.createUnitLiteral();
	      el.setName(lit.getText());
	      el.setLocationReference(this.getFilename(), lit.getLine()); }
	    { ut.addUnitLiteral(el); }
	( COMMA morelit: IDENT 
	    { el = PF.createUnitLiteral();
	      el.setName(morelit.getText());
	      el.setLocationReference(this.getFilename(), morelit.getLine()); }
	  ASSIGN unitid: IDENT STAR nv = number_value  
	    { ut.addUnitLiteral(el); el.setFactor(nv);}
	    { el.setParsedUnitLiteral(unitid.getText(),this.getFilename(),unitid.getLine());}
	)* RPAREN
//-------------------------------------------------------------------
;


real_type [PropertySet ps, Token il]
		{ Aadlreal ar;LocationReference lr=null;}:
	ar = unnamed_real_type
	  {  ar.setName(il.getText()); ps.addPropertyType(ar);processComments(ar);}
         {  lr = new LocationReference(this.getFilename(),il.getLine()); ar.setLocationReference(lr);}
;    

unnamed_real_type returns [Aadlreal ar= PF.createAadlreal()]
		{ ParsedPropertyReference ppr = null; UnitsType ut = null;}:
    REAL 
    ( real_range[ar] )?   
    ( UNITS 
      ( ppr = unique_property_type_identifier
	        { ar.setParsedPropertyReference(ppr);}
       | { ut = PF.createUnitsType(); ar.setUnitsType(ut);}
         units_list[ut] 
      )
     )?
;    

integer_type [PropertySet ps, Token il]
		{ Aadlinteger ai;LocationReference lr=null;}:
	ai = unnamed_integer_type
	  {  ai.setName(il.getText()); ps.addPropertyType(ai);processComments(ai);}
         {  lr = new LocationReference(this.getFilename(),il.getLine()); ai.setLocationReference(lr);}
;    

unnamed_integer_type returns [Aadlinteger ai= PF.createAadlinteger()]
		{ ParsedPropertyReference ppr = null;UnitsType ut= null;}:
	i:INTEGER { ai.setLocationReference(this.getFilename(), i.getLine()); }
	(integer_range[ai] )? 
	( UNITS 
	  (ppr =  unique_property_type_identifier
	        { ai.setParsedPropertyReference(ppr);}
	    | { ut = PF.createUnitsType(); ai.setUnitsType(ut);}
	    units_list[ut] 
	  )
	)?
;

// For these next rules we know that we have only reals or only integers due to a preceding reserverd word
// We also know that any identiifier is a reference to a property constant of the appropriate type.
// Those references are recorded as ParsedPropertyReferences with the RealValue and IntegerValue instead of the generic reference class

real_range [Aadlreal ar]
        { NumberOrPropertyReference bound = null;}:
    bound = signed_aadlreal_or_constant { ar.setLower(bound);}
    DOTDOT bound = signed_aadlreal_or_constant { ar.setUpper(bound);}
//-------------------------------------------------------------------
;

integer_range [Aadlinteger ai]
        { NumberOrPropertyReference bound = null;}:
    bound = signed_aadlinteger_or_constant { ai.setLower(bound);}
    DOTDOT 
	bound = signed_aadlinteger_or_constant { ai.setUpper(bound);}
//-------------------------------------------------------------------
;

// real literal with an optional sign and unit
signed_aadlreal returns [RealValue cv=PF.createRealValue()]
	{ char sign = ' '; Token numericval;}
  :
    ( PLUS {sign = '+';}
    | MINUS {sign = '-';} 
    )?  
    ( numericval = real_literal   
        { cv.setValueString(sign == ' '?numericval.getText():sign+numericval.getText()); 
          cv.setLocationReference(this.getFilename(), numericval.getLine()); }
    )
    // optional unit
       ( ui:IDENT 
          { cv.setParsedUnitLiteral(ui.getText(),this.getFilename(),ui.getLine()); }
       )?
    ;


// real literal with optional sign, named reference with or without sign
// Note that here we know that an identifier is referring to a real
signed_aadlreal_or_constant returns [NumberOrPropertyReference cv=null]
       {  String sign =null;}
   :
       cv = signed_aadlreal
    |
    (
        ( PLUS {sign = "+";}
        | MINUS {sign = "-";} 
        )? 
        cv = property_name_constant_reference 
    	{ if ( sign != null) ((PropertyReference)cv).setReferenceSign(sign);
    	}
    )
;

// integer literal with optional sign and unit
signed_aadlinteger returns [IntegerValue cv=PF.createIntegerValue()]
	{ char sign = ' '; Token numericval;}
  :
    ( PLUS {sign = '+';}
    | MINUS {sign = '-';} 
    )?  
    ( 
    numericval = integer_literal 
       { cv.setValueString(sign == ' '?numericval.getText():sign+numericval.getText()); 
         cv.setLocationReference(this.getFilename(), numericval.getLine()); }
    )
    // optional unit
       ( ui:IDENT 
          { cv.setParsedUnitLiteral(ui.getText(),this.getFilename(),ui.getLine()); }
       )?
    ;

// integer literal with optional sign and unit, named reference with or without sign
// Note that here we know that an identifier is referring to a integer
signed_aadlinteger_or_constant returns [NumberOrPropertyReference cv=null]
	{ String sign = null;}
  :
	cv = signed_aadlinteger
    | 
    (
        ( PLUS {sign = "+";}
        | MINUS {sign = "-";} 
        )? 
	cv = property_name_constant_reference 
    	{ if (sign != null) ((PropertyReference)cv).setReferenceSign(sign);
    	}
    )
;	


// numeric literal with optional sign and unit
signed_aadlnumeric returns [NumberValue cv=null]
	{ char sign = ' '; }
  :
    ( PLUS {sign = '+';}
    | MINUS {sign = '-';} 
    )?  
    ( 
    numericval: NUMERIC_LIT 
       { String str = numericval.getText();
         if ( str.indexOf('.') == -1 ) 
         	cv =PF.createIntegerValue();
         else
         	cv =PF.createRealValue();
       	
       	cv.setValueString(sign == ' '?str:sign+str); 
         cv.setLocationReference(this.getFilename(), numericval.getLine()); }
    )
    // optional unit
       ( ui:IDENT 
          { cv.setParsedUnitLiteral(ui.getText(),this.getFilename(),ui.getLine()); }
       )?
    ;



// numeric literal without optional sign and unit
number_value returns [NumberValue cv=null]
	{ char sign = ' '; }
  :
    ( 
    numericval: NUMERIC_LIT 
       { String str = numericval.getText();
         if ( str.indexOf('.') == -1 ) 
         	cv =PF.createIntegerValue();
         else
         	cv =PF.createRealValue();
       	
       	cv.setValueString(sign == ' '?str:sign+str); 
         cv.setLocationReference(this.getFilename(), numericval.getLine()); }
    )
    ;

// real literal with optional sign, named reference with or without sign
// Note that here we know that an identifier is referring to a real
signed_aadlnumeric_or_constant returns [NumberOrPropertyReference cv=null]
       { ParsedPropertyReference ppr = null; String sign =null;}
   :
   	cv = signed_aadlnumeric
   	|
   	(
        ( PLUS {sign = "+";}
        | MINUS {sign = "-";} 
        )? 
        cv = property_name_constant_reference 
    	{ if(sign != null) ((PropertyReference)cv).setReferenceSign(sign);
    	}
    )
;


// real literal with optional sign, named reference with or without sign
// Note that here we know that an identifier is referring to a real
signed_aadlnumeric_or_signed_constant returns [PropertyValue cv=null]
       { ParsedPropertyReference ppr = null; String sign =null;}
   :
   	cv = signed_aadlnumeric
   	|
   	(
        ( PLUS {sign = "+";}
        | MINUS {sign = "-";} 
        ) 
        cv = property_name_constant_reference 
    	{ if(sign != null) ((PropertyReference)cv).setReferenceSign(sign);
    	}
    )
;


// note that in case of the range type any name references in the range are created in terms of Real or Integer objects with a ParsedProeprtyReference    
unnamed_range_type returns[ RangeType rt = PF.createRangeType()] 
        { ParsedPropertyReference ppr = null; Aadlreal urt = null;Aadlinteger uit = null;}:
	RANGE OF 
	( urt = unnamed_real_type { rt.setPropertyType(urt);}
	| uit = unnamed_integer_type { rt.setPropertyType(uit);}
	| ppr = unique_property_type_identifier 
    	{ rt.setParsedPropertyReference(ppr);}
	)
	;

range_type [PropertySet ps, Token il]
		{RangeType rt = PF.createRangeType();LocationReference lr=null;}:
	rt = unnamed_range_type
	  { rt.setName(il.getText()); ps.addPropertyType(rt);}
         {  lr = new LocationReference(this.getFilename(),il.getLine()); rt.setLocationReference(lr);}
;

unnamed_classifier_type returns [ClassifierType ct = PF.createClassifierType()]
    	{ ComponentCategory cc; }:
    CLASSIFIER
	( LPAREN cc= component_category 
		{ct.addCategory(cc);}
	( id: COMMA cc = component_category 
		{
			if ( ct.getCategory().contains(cc))
				errReporter.warning(this.getFilename(), id.getLine(),"Duplicate component category "+cc.getReadableName());
			else
				ct.addCategory(cc);
		}
	)* RPAREN )?
	;
	
classifier_type [PropertySet ps, Token il]
		{ ClassifierType ct = null;LocationReference lr=null;}:
	ct = unnamed_classifier_type
	  { ct.setName(il.getText()); ps.addPropertyType(ct);}
         {  lr = new LocationReference(this.getFilename(),il.getLine()); ct.setLocationReference(lr);}
;

unnamed_component_property_type returns [ ReferenceType ct = PF.createReferenceType()]
    	{ ReferableElementCategory cc; }:
    REFERENCE
	( LPAREN cc = referable_element_category
		{ct.addCategory(cc);}
	( id:COMMA cc = referable_element_category 
		{
			if ( ct.getCategory().contains(cc))
				errReporter.warning(this.getFilename(),id.getLine(),"Duplicate referable element category "+cc.getReadableName());
			else
			    ct.addCategory(cc);
		}
	)* RPAREN )?
	;
	
	
component_property_type [PropertySet ps, Token il]
		{ReferenceType ct =null;LocationReference lr=null;}:
	ct = unnamed_component_property_type
	  { ct.setName(il.getText()); ps.addPropertyType(ct);}
         {  lr = new LocationReference(this.getFilename(),il.getLine()); ct.setLocationReference(lr);}
;

referable_element_category returns [ReferableElementCategory cat=null]:
    ( THREAD GROUP { cat = ReferableElementCategory.THREAD_GROUP_LITERAL; }
    | THREAD { cat = ReferableElementCategory.THREAD_LITERAL; }
    | DATA { cat = ReferableElementCategory.DATA_LITERAL; }
    | SUBPROGRAM { cat = ReferableElementCategory.SUBPROGRAM_LITERAL; }
    | PROCESS { cat = ReferableElementCategory.PROCESS_LITERAL; }
    | PROCESSOR { cat = ReferableElementCategory.PROCESSOR_LITERAL; }
    | MEMORY { cat = ReferableElementCategory.MEMORY_LITERAL; }
    | BUS { cat = ReferableElementCategory.BUS_LITERAL; }
    | DEVICE { cat = ReferableElementCategory.DEVICE_LITERAL; }
    | SYSTEM { cat = ReferableElementCategory.SYSTEM_LITERAL; } 
    | CONNECTIONS { cat = ReferableElementCategory.CONNECTIONS_LITERAL; }
    | SERVER SUBPROGRAM { cat = ReferableElementCategory.SERVER_SUBPROGRAM_LITERAL; }
     )
    ; 

unique_property_type_identifier returns [ ParsedPropertyReference pr= new ParsedPropertyReference()]:
    ( pset: IDENT 
    {
    	pr.setPropertySetName(pset.getText());
    	referencedPropertySets.add(pset.getText());
    }
    DOUBLECOLON )? ptype: IDENT { pr.setPropertyName(ptype.getText());pr.setFilename(this.getFilename());pr.setLine(ptype.getLine());}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 5.1.2 Property Names
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
;
property_name_declaration [PropertySet ps, Token il]
options {defaultErrorHandler=true;}
        {  PropertyDefinition pn = PF.createPropertyDefinition();processComments(pn);
	   pn.setName(il.getText()); ps.addPropertyDefinition(pn);
          LocationReference lr = new LocationReference(this.getFilename(),il.getLine()); 
          pn.setLocationReference(lr);}:
    ( ACCESS {pn.setAccess(true);} )?
	( INHERIT {pn.setInherit(true);} )?
	( single_valued_property[pn] | multi_valued_property[pn] )
	APPLIES TO
	    { PropertyOwnerCategory pc; }
	LPAREN 
	( (  property_owner[pn]
	    ( COMMA property_owner[pn] 
	    )* 
	   ) 
	   | ( ALL 
	   // set all values by hand
	   {		for (Iterator i = PropertyOwnerCategory.VALUES.iterator(); i.hasNext();) {
			        pn.addAppliesto( (PropertyOwnerCategory)i.next());
	            }
	   }
	) )
	RPAREN SEMI
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


property_owner [PropertyDefinition pd]
	    { PropertyOwnerCategory cat=null; int lineno = 0;
	    ClassifierReference cr = null; }:
    ( id1:THREAD GROUP { cat = PropertyOwnerCategory.THREAD_GROUP_LITERAL; lineno = id1.getLine();}
    | id2:THREAD { cat = PropertyOwnerCategory.THREAD_LITERAL; lineno = id2.getLine();}
    | id3:DATA { cat = PropertyOwnerCategory.DATA_LITERAL; lineno = id3.getLine();}
    | id4:SUBPROGRAM { cat = PropertyOwnerCategory.SUBPROGRAM_LITERAL; lineno = id4.getLine();}
    | id5:PROCESS { cat = PropertyOwnerCategory.PROCESS_LITERAL; lineno = id5.getLine();}
    | id6:PROCESSOR { cat = PropertyOwnerCategory.PROCESSOR_LITERAL; lineno = id6.getLine();}
    | id7:MEMORY { cat = PropertyOwnerCategory.MEMORY_LITERAL; lineno = id7.getLine();}
    | id8:BUS { cat = PropertyOwnerCategory.BUS_LITERAL; lineno = id8.getLine();}
    | id9:DEVICE { cat = PropertyOwnerCategory.DEVICE_LITERAL; lineno = id9.getLine();}
    | id10:SYSTEM { cat = PropertyOwnerCategory.SYSTEM_LITERAL; lineno = id10.getLine();} 
    | id11:CONNECTIONS { cat = PropertyOwnerCategory.CONNECTIONS_LITERAL; lineno = id11.getLine();}
	| id12:SERVER SUBPROGRAM { cat = PropertyOwnerCategory.SERVER_SUBPROGRAM_LITERAL; lineno = id12.getLine();}
	 | id13:MODE { cat = PropertyOwnerCategory.MODE_LITERAL; lineno = id13.getLine();}
	 | id14:PORT GROUP CONNECTIONS { cat = PropertyOwnerCategory.PORT_GROUP_CONNECTIONS_LITERAL; lineno = id14.getLine();}
	 | id15:PORT GROUP { cat = PropertyOwnerCategory.PORT_GROUP_LITERAL; lineno = id15.getLine();}
	 | id16:PORT CONNECTIONS { cat = PropertyOwnerCategory.PORT_CONNECTIONS_LITERAL; lineno = id16.getLine();}
	 | id17:PORT { cat = PropertyOwnerCategory.PORT_LITERAL; lineno = id17.getLine();}
	 | id18:DATA PORT { cat = PropertyOwnerCategory.DATA_PORT_LITERAL; lineno = id18.getLine();}
	 | id19:EVENT DATA PORT { cat = PropertyOwnerCategory.EVENT_DATA_PORT_LITERAL; lineno = id19.getLine();}
	 | id20:EVENT PORT { cat = PropertyOwnerCategory.EVENT_PORT_LITERAL; lineno = id20.getLine();}
	 | id21:DATA PORT CONNECTIONS { cat = PropertyOwnerCategory.DATA_PORT_CONNECTIONS_LITERAL; lineno = id21.getLine();}
	 | id22:EVENT DATA PORT CONNECTIONS { cat = PropertyOwnerCategory.EVENT_DATA_PORT_CONNECTIONS_LITERAL; lineno = id22.getLine();}
	 | id23:EVENT PORT CONNECTIONS { cat = PropertyOwnerCategory.EVENT_PORT_CONNECTIONS_LITERAL; lineno = id23.getLine();}
	 | id24:FLOW { cat = PropertyOwnerCategory.FLOW_LITERAL; lineno = id24.getLine();}
	 | id25:ACCESS CONNECTIONS { cat = PropertyOwnerCategory.ACCESS_CONNECTIONS_LITERAL; lineno = id25.getLine();}
	 | id26:PARAMETER { cat = PropertyOwnerCategory.PARAMETER_LITERAL; lineno = id26.getLine();}
	 | id27:PARAMETER CONNECTIONS { cat = PropertyOwnerCategory.PARAMETER_CONNECTIONS_LITERAL; lineno = id27.getLine();}
	 )
	( cr = classifier_reference	)?
	{ 
		if ( cr == null) {
			if ( pd.getAppliesto().contains(cat) ) 
			    errReporter.warning(this.getFilename(), lineno,"Duplicate property owner category "+cat.getReadableName());
			else
				pd.addAppliesto(cat);
		} else {
			if ( cat.getValue() > PropertyOwnerCategory.SYSTEM) {
			    errReporter.warning(this.getFilename(), lineno,"No classifier for property owner category "+ cat.getReadableName());
			} else {
				ClassifierValue cv = PF.createClassifierValue();
				cv.setValue(ComponentCategory.get(cat.getValue()));
				cv.setClassifierReference(cr);
				pd.addAppliesToClassifier(cv);
			}
		}
	}
;

single_valued_property [PropertyDefinition pn]
     		{ PropertyValue val = null; ParsedPropertyReference ppr=null;
     		PropertyType upt;}:
    ( 
      upt = unnamed_property_type 
      	{ pn.setPropertyType(upt);}
     | 
        ppr = unique_property_type_identifier
            { pn.setParsedPropertyReference(ppr); }
    )
    
    ( ASSIGN val = property_expression 
    	{  pn.addDefaultpropertyValue(val); }
    )?
;


multi_valued_property [PropertyDefinition pn]
     		{ PropertyValue val = null; ParsedPropertyReference ppr=null;
     		PropertyType upt;}:
	LIST OF {pn.setList(true);} 
    ( 
      upt = unnamed_property_type 
      	{ pn.setPropertyType(upt);}
     | 
        ppr = unique_property_type_identifier
            { pn.setParsedPropertyReference(ppr); }
    )
    
    ( ASSIGN LPAREN { pn.setHasEmptyList(true); }
    ( val = property_expression 
    		{  pn.addDefaultpropertyValue(val); pn.setHasEmptyList(false);} 
        ( COMMA val = property_expression { pn.addDefaultpropertyValue(val);}
        )*
    )?
    RPAREN 
    )?
;



// Note that for property constants we know we are dealing with reals and integers,
// even when identifiers are used - they refer to other constants
// As a result we create the respective Value objects and attach the ParsedPropertyReference object to it.


property_constant [PropertySet ps, Token il]
options {defaultErrorHandler=true;}
		{ PropertyConstant pc = PF.createPropertyConstant();processComments(pc);
		 pc.setName(il.getText());ps.addPropertyConstant(pc);
          LocationReference lr = new LocationReference(this.getFilename(),il.getLine()); 
          pc.setLocationReference(lr);}:
     CONSTANT  
     (single_valued_property_constant[pc]
     |
     multi_valued_property_constant[pc])
     SEMI
     ;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

     
 single_valued_property_constant[PropertyConstant pc] 
 				{ ParsedPropertyReference ppr = null; 
 				PropertyValue val;  Aadlinteger ai = null;
 				Aadlreal ar=null;
 				}
 :
    (( INTEGER 
     			{ ai = PF.createAadlinteger();
     				pc.setPropertyType(ai);
     			}
     	(
    		ppr = unique_property_type_identifier  
    				{ ai.setParsedPropertyReference(ppr); }    	
     	)?
       ASSIGN 
        val = signed_aadlinteger { pc.addConstantValue(val); }
       )
      |( REAL 
     			{  ar = PF.createAadlreal();
     				pc.setPropertyType(ar);
     			}
          (
     		ppr = unique_property_type_identifier  
     				{ ar.setParsedPropertyReference(ppr); }    	
          )?
         ASSIGN 
        val = signed_aadlreal { pc.addConstantValue(val); }
      )
     |( STRING ASSIGN 
       { pc.setPropertyType(PF.createAadlstring()); }
       val = string_term 
       { pc.addConstantValue(val); }
     )
     |( BOOLEAN ASSIGN 
       { pc.setPropertyType(PF.createAadlboolean()); }
       val = true_false_value 
       { pc.addConstantValue(val); }
     )
     |( 
     	{  ppr = null; }
     ppr = unique_property_type_identifier ASSIGN 
     		{ pc.setParsedPropertyReference(ppr); }
       (val = enumeration_term { pc.addConstantValue(val); }
        | 
        (
        	( (signed_aadlnumeric DOTDOT) =>
         	val = aadl_numeric_range 
         	|
       		val = signed_aadlnumeric
       		)
	       	{ pc.addConstantValue(val); }
       	)
       )
     )
    ) 
    ;


     
 multi_valued_property_constant[PropertyConstant pc] 
     		{ ParsedPropertyReference ppr = null; 
     			 PropertyValue val; Aadlinteger ai = null;
 				Aadlreal ar=null;
     		}
 :
     LIST OF {pc.setList(true);}
    (( INTEGER 
     			{  ai = PF.createAadlinteger();
     				pc.setPropertyType(ai);}
    	(
    		ppr = unique_property_type_identifier  
    				{ ai.setParsedPropertyReference(ppr); }    	
     	)?
    ASSIGN LPAREN
      ( val = signed_aadlinteger { pc.addConstantValue(val); }
        ( COMMA val = signed_aadlinteger { pc.addConstantValue(val); }
        )* 
      )?
      RPAREN
     )
    |( REAL 
     			{  ar = PF.createAadlreal();
     				pc.setPropertyType(ar);
     			}
          (
     		ppr = unique_property_type_identifier  
     				{ ar.setParsedPropertyReference(ppr); }    	
          )?
       ASSIGN LPAREN
      ( val = signed_aadlreal { pc.addConstantValue(val); }
        ( COMMA val = signed_aadlreal { pc.addConstantValue(val); }
        )*
      )?
      RPAREN
     )
     |( STRING 
     		{ pc.setPropertyType(PF.createAadlstring()); }
        ASSIGN LPAREN
      ( val = string_term { pc.addConstantValue(val); }
        ( COMMA val = string_term { pc.addConstantValue(val); }
        )*
      )?
      RPAREN
     )
     |( BOOLEAN 
     		{ pc.setPropertyType(PF.createAadlboolean()); }
        ASSIGN LPAREN
       ( val = true_false_value { pc.addConstantValue(val); }
         ( COMMA val = true_false_value { pc.addConstantValue(val); }
         )*
        )?
        RPAREN
     )
     |( 
     	{  ppr = null; }
     	ppr = unique_property_type_identifier ASSIGN LPAREN
     		{ pc.setParsedPropertyReference(ppr); }
       ( ( val = enumeration_term { pc.addConstantValue(val); }
         	( COMMA val = enumeration_term { pc.addConstantValue(val); }
         	)*
         )
         |
          ( (signed_aadlnumeric DOTDOT) =>
         	( val = aadl_numeric_range { pc.addConstantValue(val); }
         	  ( COMMA val = aadl_numeric_range { pc.addConstantValue(val); }
         	  )*
             )
             |
       		 ( val = signed_aadlnumeric { pc.addConstantValue(val); }
         	  ( COMMA val = signed_aadlnumeric { pc.addConstantValue(val); }
         	  )*
       		 )
       		)  // signed_aadlnumeric DOTDOT
       	)?
       	RPAREN
     )
   )
   ;

aadl_numeric_range returns [RangeValue nrv=null]
	{ NumberValue pv = null; /*RangeValue nrv = null;*/ NumberValue val = null; }
:
    pv = signed_aadlnumeric
	DOTDOT val = signed_aadlnumeric
	{
			if ( pv instanceof IntegerValue ){
	     		nrv = PF.createIntegerRangeValue();
	     		((IntegerRangeValue)nrv).setMinimum((IntegerValue)pv); 
	        } else if ( pv instanceof RealValue ){
	     		nrv = PF.createRealRangeValue();
	     		((RealRangeValue)nrv).setMinimum((RealValue)pv);
		    }
		    if (nrv != null) nrv.setLocationReference(pv.getLocationReference());
	   		if ( nrv instanceof IntegerRangeValue) {
	   			if ( val instanceof IntegerValue){
	   			    ((IntegerRangeValue)nrv).setMaximum((IntegerValue)val); 
	   			} else {
	   				// expected integer max value, but got real value
	   				errReporter.warning(nrv.getLocationReference(),"Expected integer maximum value, but got real maximum");
	   			}
	   		} else if ( nrv instanceof RealRangeValue ) {
	   			if ( val instanceof RealValue){
	   			    ((RealRangeValue)nrv).setMaximum((RealValue)val); 
	   			} else {
	   				// expected real max value, but got integer value
	   				errReporter.warning(nrv.getLocationReference(),"Expected real maximum value, but got integer maximum");
	   			}
	   		}
	}
	( DELTA val = signed_aadlnumeric
	{
	   		if ( nrv instanceof IntegerRangeValue) {
	   			if ( val instanceof IntegerValue){
	   			    ((IntegerRangeValue)nrv).setDelta((IntegerValue)val); 
	   			} else {
	   				// expected integer max value, but got real value
	   				errReporter.warning(nrv.getLocationReference(),"Expected integer delta value, but got real delta");
	   			}
	   		} else if ( nrv instanceof RealRangeValue ) {
	   			if ( val instanceof RealValue){
	   			    ((RealRangeValue)nrv).setDelta((RealValue)val); 
	   			} else {
	   				// expected real max value, but got integer value
	   				errReporter.warning(nrv.getLocationReference(),"Expected real delta value, but got integer delta");
	   			}
	   		}
	}
	)?
;

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 5.3 Property Associations
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

propertyAssociations returns [Properties ps = null]
//options {defaultErrorHandler=true;}
{ PropertyAssociation pa ; }:
   PROPERTIES { ps = PF.createProperties(); processComments(ps);}
    (none_stmt | 
      (  pa = property_association { ps.addPropertyAssociation(pa); }
      )+ ) ;


containedPropertyAssociations returns [Properties ps = null]
//options {defaultErrorHandler=true;}
{ PropertyAssociation pa ; }:
   PROPERTIES { ps = PF.createProperties(); }
    (none_stmt | 
      (  pa = contained_property_association { ps.addPropertyAssociation(pa); processComments(pa);}
      )+ ) ;

propertyAssociations_no_modes returns [Properties ps = null]
//options {defaultErrorHandler=true;}
 { PropertyAssociation pa ; }:
   PROPERTIES { ps = PF.createProperties(); processComments(ps);}
    (none_stmt | 
      (  pa = property_association_no_modes { ps.addPropertyAssociation(pa); }
      )+ ) ;


curlyPropertyAssociations returns [Properties ps = null]
options {defaultErrorHandler=true;}
{ PropertyAssociation pa ; }:
     LCURLY { ps = PF.createProperties(); processComments(ps);}
    ( pa = property_association { ps.addPropertyAssociation(pa); }
    )+ RCURLY  ;


curlyPropertyAssociations_no_modes returns [Properties ps = null]
options {defaultErrorHandler=true;} 
{ PropertyAssociation pa ; }:
     LCURLY { ps = PF.createProperties(); processComments(ps);}
    ( pa = property_association_no_modes { ps.addPropertyAssociation(pa); }
    )+ RCURLY  ;


containedCurlyPropertyAssociations returns [Properties ps = null]
options {defaultErrorHandler=true;}
{ PropertyAssociation pa ; }:
   LCURLY { ps = PF.createProperties(); processComments(ps);}
    ( pa =  contained_property_association { ps.addPropertyAssociation(pa); }
    )+ RCURLY  ;


property_association returns [PropertyAssociation pa = PF.createPropertyAssociation()] 
//options {defaultErrorHandler=true;}
    	{ PropertyValue val = null;ParsedPropertyReference pr = null; }
:
    	{processComments(pa);}
    pr = property_name_reference { 
    		pa.setParsedPropertyReference(pr);
    		    pa.setLocationReference(this.getFilename(),pr.getLine());
    		}
    ( ASSIGN | ASSIGNPLUS { pa.setAppend(true);} )
    
    ( CONSTANT { pa.setConstant(true);} 
    )?
    ( ACCESS { pa.setAccess(true);} 
    )?
    pe_or_list[pa]
    ( in_binding[pa] )?
    ( in_modes[pa] )?
		{ if (pa.isAppend() && (!pa.getModeName().isEmpty() || pa.isNoMode())){
				errReporter.error(pa.getLocationReference(),"Append not permitted for in modes property values");			
			}
		}
    SEMI;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }



property_association_no_modes returns [PropertyAssociation pa = PF.createPropertyAssociation()] 
//options {defaultErrorHandler=true;}
    	{ PropertyValue val = null;ParsedPropertyReference pr = null; }
:
    	{processComments(pa);}
    pr = property_name_reference { 
    		pa.setParsedPropertyReference(pr);
    		pa.setLocationReference(this.getFilename(),pr.getLine()); }

    ( ASSIGN | ASSIGNPLUS { pa.setAppend(true);} )
    
    ( CONSTANT { pa.setConstant(true);} 
    )?
    ( ACCESS { pa.setAccess(true);} 
    )?
    pe_or_list[pa]

    ( in_binding[pa] )?
    SEMI;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }



property_name_reference returns [ParsedPropertyReference pr = new ParsedPropertyReference()]:
    ( psid:IDENT DOUBLECOLON 
    		{
    			pr.setPropertySetName(psid.getText());
    			referencedPropertySets.add(psid.getText());
    		} 
    )?
    pid:IDENT  { pr.setPropertyName(pid.getText());pr.setFilename(this.getFilename());pr.setLine(pid.getLine());}
;


property_name_constant_reference returns [PropertyReference pr = PF.createPropertyReference()]
		{ ParsedPropertyReference ppr;  }:
    VALUE LPAREN
    ppr = property_name_reference
    RPAREN
    	{pr.setParsedPropertyReference(ppr);
    		pr.setLocationReference(this.getFilename(),ppr.getLine());
    	}
;

// NOTE: This is not exactly the same as the contained_property_assocation rule in the
// spec.  In this case, "contained_property_assocation" is used whereever contained
// property assocations are allowed, and thus this rule really handles both normal and
// contained property associations.
contained_property_association returns [PropertyAssociation pa = PF.createPropertyAssociation()]
    	{ PropertyValue val = null;ParsedPropertyReference pr = null;processComments(pa);}:
    pr = property_name_reference { pa.setParsedPropertyReference(pr);
    pa.setLocationReference(this.getFilename(),pr.getLine()); }
    ( ASSIGN | ASSIGNPLUS { pa.setAppend(true); } )
    ( CONSTANT { pa.setConstant(true);} 
    )?

    pe_or_list[pa]
    ( applies_to[pa] )?
		{ if (pa.isAppend() && (pa.getComponentPath() != null && !pa.getComponentPath().isEmpty())){
				errReporter.error(pa.getLocationReference(),"Append not permitted for contained property associations");			
			}
		}
    ( in_binding[pa] )?
    ( in_modes[pa] )?
		{ if (pa.isAppend() && (!pa.getModeName().isEmpty() || pa.isNoMode())){
				errReporter.error(pa.getLocationReference(),"Append not permitted for in modes property values");			
			}
		}
 SEMI;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

applies_to [ PropertyAssociation pa]:
APPLIES TO cid: IDENT { pa.addComponentPath(cid.getText());}
		( DOT mid: IDENT { pa.addComponentPath(mid.getText());}
		)* 
		;

in_binding [ PropertyAssociation pa] {ClassifierReference cr;}:
	IN BINDING LPAREN cr = classifier_reference {pa.addClassifierReference(cr);}
	( COMMA cr = classifier_reference {pa.addClassifierReference(cr);}
	)*
	RPAREN
		{ if (pa.isAppend()){
				errReporter.error(pa.getLocationReference(),"Append not permitted for in binding propery values");			
			}
		}
;

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 5.4 Property Expressions
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Same as pe_or_list but callable as a start rule
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

pe_or_list_top_level returns [PropertyAssociation pa = PF.createPropertyAssociation()]
:
  pe_or_list[pa]
;

	
pe_or_list [PropertyAssociation pa]
	{ ParsedPropertyReference ppr= null;
		PropertyValue val = null;
	}
:
( LPAREN property_expression COMMA ) => 
	(LPAREN val = property_expression 
		{ pa.addPropertyValue(val);}
	( COMMA val = property_expression 
		{ pa.addPropertyValue(val);}
	)+ RPAREN
	)
	|
	( LPAREN
	( val=allbutbool )?
		{ pa.addPropertyValue(val);}
	RPAREN
	)
	| 
	(val= allbutbool 
			{ pa.addPropertyValue(val);}
	)
	|
	(
		( property_name_constant_reference DOTDOT ) =>
		( val = property_name_constant_reference
		  val = num_range[val] 
		    { pa.addPropertyValue(val); }
		)
		| 	
		( val = logical_or
			{ pa.addPropertyValue(val);}
		)
	)
;

allbutbool returns [PropertyValue pv] 
:
	 pv = reference_term
	| pv = enumeration_term
	| pv = string_term
	| pv = component_classifier_term
 	| pv = numeric_or_range_term
;



property_expression returns [PropertyValue pv = null]:
//   property_term is already covered in boolean expression (logical_or)
 	pv = numeric_or_range_term
	| pv = reference_term
	| pv = enumeration_term
	| pv = string_term
	| pv = component_classifier_term
	| pv = logical_or
	 ;


logical_or returns [ BooleanOrPropertyReference bv = null]
        { BooleanOrPropertyReference  bv1, bv2; }:
    bv = logical_and ( OR bv2 = logical_and 
         { bv1 = bv; bv = PF.createBooleanOR();
         	  ((BooleanOR)bv).addBooleanValue(bv1);
         	  ((BooleanOR)bv).addBooleanValue(bv2);
         	  bv.setLocationReference(bv1.getLocationReference()); }
    )*
    ;
    
logical_and returns [ BooleanOrPropertyReference bv = null]
        { BooleanOrPropertyReference  bv1, bv2; }:
    bv = logical_unary ( AND bv2 = logical_unary 
         { bv1 = bv; bv = PF.createBooleanAND(); 
         	  ((BooleanAND)bv).addBooleanValue(bv1);
         	  ((BooleanAND)bv).addBooleanValue(bv2);
         	  bv.setLocationReference(bv1.getLocationReference()); }
    )*
    ;
    
logical_unary returns [ BooleanOrPropertyReference bv = null]
        { BooleanOrPropertyReference  bv1; }
    :
    tv:TRUE { bv = PF.createTRUE(); bv.setLocationReference(this.getFilename(), tv.getLine()); }
    | fv:FALSE { bv = PF.createFALSE(); bv.setLocationReference(this.getFilename(), fv.getLine()); }
    | LPAREN bv = logical_or RPAREN 
    | not:NOT bv1 = logical_unary { 
    	    bv = PF.createBooleanNOT();
    	    ((BooleanNOT)bv).setBooleanValue(bv1);
    	    bv.setLocationReference(this.getFilename(), not.getLine());
    	  }
    | bv = property_term
    ;

true_false_value returns [ TrueFalseValue bv = null]
    :
    tv:TRUE { bv = PF.createTRUE(); bv.setLocationReference(this.getFilename(), tv.getLine()); }
    | fv:FALSE { bv = PF.createFALSE(); bv.setLocationReference(this.getFilename(), fv.getLine()); }
;


signed_constant returns [PropertyReference pv=null]
       { ParsedPropertyReference ppr = null; String sign =null;}
   :
        ( PLUS {sign = "+";}
        | MINUS {sign = "-";} 
        )
        pv = property_name_constant_reference 
    	{ if (sign != null) pv.setReferenceSign(sign);
    	}
;

// this rule handles integers, reals, as well as ranges of both types
// This includes the use of identifiers when they are preceded by a sign
// or all identifiers in the max and delta part of ranges
// Those references are represented as IntegerValue if known, with a ParsedPropertyReference object.
// Otherwise they are represented as RealValue - it may have to be recast into an IntergerValue at the time of name resolution.



numeric_or_range_term returns [PropertyValue pv=null]
        { NumberValue val; }
	:
	( signed_constant DOTDOT ) =>
	  (	pv = signed_constant 
	  pv = num_range[pv]
	)
	|
    ( pv = signed_aadlnumeric_or_signed_constant
		 (  	pv = num_range[pv]
		 )?
	   )
         ;

// support rule to handle remainder of range if DOTDOT is identified in look ahead
num_range [PropertyValue pv] returns [RangeValue nrv=null]
{ NumberOrPropertyReference val = null;}:
			{ if ( pv instanceof IntegerValue ){
		     		nrv = PF.createIntegerRangeValue();
		     		((IntegerRangeValue)nrv).setMinimum((IntegerValue)pv); 
		      } else if ( pv instanceof RealValue ){
		     		nrv = PF.createRealRangeValue();
		     		((RealRangeValue)nrv).setMinimum((RealValue)pv);
		      } else if ( pv instanceof PropertyReference){
		     		nrv = PF.createRangeValue();
	   		  	((RangeValue)nrv).setMinimum((PropertyReference)pv);
			  }
			  LocationReference loc = pv.getLocationReference();
			  if ( loc != null) 
			  	nrv.setLocationReference(loc.getFilename(),loc.getLine());
		  	}
		   DOTDOT val = signed_aadlnumeric_or_constant 
		   		{ nrv.setMaximum(val); 
		   		  if (nrv instanceof IntegerRangeValue && val instanceof RealValue) {
		   		  	errReporter.warning(nrv.getLocationReference(), 
		   		  	  "Expected integer maximum value, but got real maximum");
		   		  }
		   		  if (nrv instanceof RealRangeValue && val instanceof IntegerValue) {
		   		  	errReporter.warning(nrv.getLocationReference(), 
		   		  	  "Expected real maximum value, but got integer maximum");
		   		  }
		   		}
		   ( DELTA val = signed_aadlnumeric_or_constant 
		   		{ nrv.setDelta(val);
		   		  if (nrv instanceof IntegerRangeValue && val instanceof RealValue) {
		   		  	errReporter.warning(nrv.getLocationReference(), 
		   		  	  "Expected integer delta value, but got real delta");
		   		  }
		   		  if (nrv instanceof RealRangeValue && val instanceof IntegerValue) {
		   		  	errReporter.warning(nrv.getLocationReference(), 
		   		  	  "Expected real delta value, but got integer delta");
		   		  }
		   		}
		   )?
;


string_term returns [ StringValue sv =PF.createStringValue()] :
     sl: STRING_LITERAL 
         { String s = sl.getText();
         	if (s.startsWith("\"")){
         		s = s.substring(1);
         	}
         	if (s.endsWith("\"")){
         		s = s.substring(0,s.length()-1);
         	}
         	sv.setValue(s); sv.setLocationReference(this.getFilename(),sl.getLine()); }
//-------------------------------------------------------------------
;

enumeration_term returns [ EnumValue ev = null]:
     enum_id: IDENT  { ev = PF.createEnumValue(); ev.setEnumLiteralName(enum_id.getText()); 
     				 ev.setParsedEnumLiteral(enum_id.getText(),this.getFilename(),enum_id.getLine());
     				 ev.setLocationReference(this.getFilename(),enum_id.getLine());
     				 }
//-------------------------------------------------------------------
;

property_term returns [ PropertyReference pr = null]:
        pr = property_name_constant_reference
//------------------------------------------------------------------
;



component_classifier_term returns [ ClassifierValue cv=null]
{ ClassifierReference cr = null; ComponentCategory cat;}:
      cat = component_category ( cr = classifier_reference )?
         { cv = PF.createClassifierValue(); cv.setClassifierReference(cr); cv.setValue(cat);
           if ( cr != null) cv.setLocationReference(this.getFilename(),cr.getLine());
         }
     ;

component_category returns [ComponentCategory cat = null]:
    ( THREAD GROUP { cat = ComponentCategory.THREAD_GROUP_LITERAL; })
    | THREAD { cat = ComponentCategory.THREAD_LITERAL; }
    | DATA { cat = ComponentCategory.DATA_LITERAL; }
    | SUBPROGRAM { cat = ComponentCategory.SUBPROGRAM_LITERAL; }
    | PROCESS { cat = ComponentCategory.PROCESS_LITERAL; }
    | PROCESSOR { cat = ComponentCategory.PROCESSOR_LITERAL; }
    | MEMORY { cat = ComponentCategory.MEMORY_LITERAL; }
    | BUS { cat = ComponentCategory.BUS_LITERAL; }
    | DEVICE { cat = ComponentCategory.DEVICE_LITERAL; }
    | SYSTEM { cat = ComponentCategory.SYSTEM_LITERAL; } 
//-------------------------------------------------------------------
;
// ** may be the same as the instance association
reference_term returns [ ReferenceValue ct = PF.createReferenceValue()]:
     REFERENCE subcomponentid: IDENT 
         { ct.addSubcomponentName(subcomponentid.getText()); }
         { LocationReference lr = new LocationReference(this.getFilename(),subcomponentid.getLine()); ct.setLocationReference(lr);}
     ( DOT subcomponentidmore: IDENT 
         { ct.addSubcomponentName(subcomponentidmore.getText()); }
     )*
//-------------------------------------------------------------------


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 6.2 Subprograms
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
;

callsSubclause returns [CallSequences cs = CF.createCallSequences() ]
        { CallSequence scs = null; processComments(cs);}:
     CALLS  
			( ( scs = subprogram_call_sequence { cs.addCallSequence(scs); }
			   )+ |  none_stmt  )
	;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(RCURLY);
			consumeUntil(SEMI);
			consume();
        }


subprogram_call_sequence returns [CallSequence cs = CF.createCallSequence()] 
        { SubprogramSubcomponent spc = null;processComments(cs);}:
    ( id: IDENT COLON 
    	{ cs.setName(id.getText()); }
    )?
    LCURLY ( spc=subprogram_call { cs.addCall(spc); }
    )+ RCURLY
    ( in_modes[cs] )? 
	SEMI
//-------------------------------------------------------------------
;
subprogram_call returns [SubprogramSubcomponent spc=CF.createSubprogramSubcomponent()] 
        { Properties ps; }:
    defcallid:IDENT COLON { spc.setName(defcallid.getText());}
    SUBPROGRAM 
    called_subprogram[spc]  
	     { processComments(spc); }
	( ps = curlyPropertyAssociations { spc.setProperties(ps);} )? 
	SEMI 
//-------------------------------------------------------------------
;
called_subprogram [SubprogramSubcomponent sc]
        { ClassifierReference cr;}:
// 	data_identifier DOT data_subprogram_identifier | unique_subprogram_classifier_reference
    cr = classifier_reference  { sc.setClassifierReference(cr); }
//-------------------------------------------------------------------
;

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 8.2 Modes
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


mode_subclause [ComponentImpl ci ]
{Modes ms=null;}:		
    MODES  {  ms = CoreF.createModes(); ci.setModes(ms);  processComments(ms); }
  (    (  mode[ms])* ( mode_transition [ms]  )*  
  |  none_stmt )
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


mode [Modes ms]
        { Properties ps=null; Mode m = null;
         ModeState mstate = ModeState.NORMAL_LITERAL; }:
     id:IDENT COLON 
         { m = CoreF.createMode();
         	  m.setName(id.getText());
         	  m.setLocationReference(new LocationReference(this.getFilename(),id.getLine()));
         	  ms.addMode(m);
         	  processComments(m); }
     ( INITIAL { mstate = ModeState.INITIAL_LITERAL;} )?
     ( REFINED TO 
        { m.setRefined(true);
           RefinedReference fr = new RefinedReference(); 
           fr.setRefinedName(id.getText()); 
           fr.setFilename(this.getFilename());fr.setLine(id.getLine());
           m.setRefinedReference(fr);  }
     )?
     MODE  
     	{m.setState(mstate);}
	 ( ps = curlyPropertyAssociations {m.setProperties(ps);} )?
     SEMI ;


mode_transition [Modes ms]
     	{ ModeTransition mt = CoreF.createModeTransition(); ms.addModeTransition(mt);processComments(mt);
     	 Vector il = new Vector();
     	 FeatureReference fr=null;
     	 LocationReference lr=null;ModeTransition cp;}:
// V2 Named mode transitions
    ( mtid:IDENT COLON  { mt.setName(mtid.getText());})?
     id: IDENT { mt.setSrcModeName(id.getText()); }
// next 3 line to be removed for V2
//     ( COMMA moreid: IDENT
//     	{ il.add(moreid.getText());}
//     )*
// end V2 removal
	 LTRANS fr=unique_port_identifier 
	 	{ mt.addPortReference(fr); }
	 ( COMMA fr= unique_port_identifier 
	 	{ mt.addPortReference(fr); }
	 )* RTRANS
	 destination_mode:IDENT { mt.setDstModeName(destination_mode.getText());}
	 			{  lr = new LocationReference(this.getFilename(),destination_mode.getLine());}
	 			{ mt.setLocationReference(lr); }
         { for ( int i = 0; i < il.size(); i++) {  cp = (ModeTransition)EcoreUtil.copy(mt); cp.setSrcModeName((String)il.get(i)); cp.setDstModeName(destination_mode.getText()); ms.addModeTransition(cp); Vector pp = mt.getPortReference(); for (int j = 0; j<pp.size();j++) cp.addPortReference((FeatureReference) pp.get(j)); }}
     SEMI ;



unique_port_identifier returns [FeatureReference cr = new FeatureReference()]:
 	compid:IDENT {cr.setCompName(compid.getText());}
 	DOT portid:IDENT {cr.setFeatureName(portid.getText());cr.setFilename(this.getFilename());cr.setLine(portid.getLine());}
   |
   soleportid:IDENT {cr.setFeatureName(soleportid.getText());cr.setCompName("");cr.setFilename(this.getFilename());cr.setLine(soleportid.getLine());}
  ;
  
 
// phf *** details for rule above: unique_port_identifier :
//    component_type_port_identifier
//	| subcomponent_identifier "." port_identifier
//  | component_type_port_group_identifier "." port_identifier


in_modes [ModeMember mm]{ LocationReference lr=null;}:
    IN MODES
    LPAREN 
    ( 
    	( m1:IDENT { mm.addModeName(m1.getText());}
            {  lr = new LocationReference(this.getFilename(),m1.getLine());}
            	{ mm.setLocationReference(lr);}
    		( COMMA m2:IDENT { mm.addModeName(m2.getText());}
    		)* 
    	) 
    	| NONE { mm.setNoMode(true); }
    ) RPAREN;
    
in_modes_and_transitions [Connection mm]:
    IN MODES
    LPAREN ( 
    	( mode_or_transition[mm] 
    		( COMMA mode_or_transition[mm] )* 
    	) 
    	| NONE { mm.setNoMode(true); }
    ) RPAREN // extra parens
;

mode_or_transition [Connection mm]{ LocationReference lr=null;}:
    mode_id: IDENT { mm.addModeName(mode_id.getText());}
            {  lr = new LocationReference(this.getFilename(),mode_id.getLine());}
            { mm.setLocationReference(lr);}
    |
    LPAREN old_mode_id:IDENT { mm.addOldModeName(old_mode_id.getText());}
    ARROW new_mode_id:IDENT { mm.addNewModeName(new_mode_id.getText());}
            {  lr = new LocationReference(this.getFilename(),new_mode_id.getLine());}
            { mm.setLocationReference(lr);}
    RPAREN
//-------------------------------------------------------------------
;




//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 9 FEATURES AND SHARED ACCESS
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

// specific features are called in category specific provides clauses
//feature :
//    port_spec | server_subprogram | subprogram_spec |
//    subcomponent_access ;
// phf: the subrules are used selectively

//feature_refinement returns [Feature ft=null]:
//    ft = port_refinement |
//    ft = server_subprogram_refinement | ft = data_subprogram_refinement |
//    ft = subcomponent_access_refinement 
//    ;
// phf: refines is handled as part of the regular rule
			
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 9.1 Ports
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
port_spec returns[Port prt=null]
{Properties ps; PortDirection pd = null; }:
     ( IN OUT { pd = PortDirection.INOUT_LITERAL; }
     | IN { pd = PortDirection.IN_LITERAL; }
     | OUT { pd = PortDirection.OUT_LITERAL; }
     ) 
     prt = port_type 
           { prt.setDirection(pd);   }
	 ( ps = curlyPropertyAssociations_no_modes  { prt.setProperties(ps);} )?
	 SEMI ;

	 
port_type returns [Port pt = null]
{ ClassifierReference cr = null;}
:
    (EVENT DATA PORT { pt = FF.createEventDataPort();processComments(pt);}
        ( cr = classifier_reference {pt.setClassifierReference(cr);})?
    ) 
    |
    (DATA PORT { pt = FF.createDataPort();processComments(pt);}
        ( cr = classifier_reference {pt.setClassifierReference(cr);})?
    ) 
    | 
    (EVENT PORT   { pt = FF.createEventPort();processComments(pt);}
    ) 
    ;
    

subprogramPort returns [Port ep=null]
       {Properties ps; ClassifierReference cr = null; }:
      (  OUT  EVENT PORT 
           { ep = FF.createEventPort(); ep.setDirection(PortDirection.OUT_LITERAL);processComments(ep);}
     | OUT EVENT DATA PORT ( cr = classifier_reference )?
           { ep = FF.createEventDataPort();  processComments(ep); }
           { ep.setDirection(PortDirection.OUT_LITERAL); ep.setClassifierReference(cr);}
     )
	 ( ps = curlyPropertyAssociations_no_modes { ep.setProperties(ps);}  )?
	 SEMI
//-------------------------------------------------------------------
;


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 9.2 Subprograms
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
data_subprogram returns [Subprogram sb=FF.createSubprogram()]
		{Properties ps; ClassifierReference cr; processComments(sb);}:
     SUBPROGRAM
	 ( cr = classifier_reference {sb.setClassifierReference(cr);} )?
	 (ps = curlyPropertyAssociations_no_modes { sb.setProperties(ps);})?
	  SEMI
//-------------------------------------------------------------------
;

server_subprogram returns [ServerSubprogram sp= FF.createServerSubprogram()]
        {Properties ps; ClassifierOrFeatureReference cr; processComments(sp);}:
     SERVER 
     SUBPROGRAM
	 ( cr = unique_subprogram_reference {sp.setClassifierOrFeatureReference(cr);} )?
	 (ps = curlyPropertyAssociations_no_modes { sp.setProperties(ps);})?
	 SEMI
//-------------------------------------------------------------------
;


unique_subprogram_reference returns [ClassifierOrFeatureReference cr = new ClassifierOrFeatureReference()] 
{String pn = "";}:
	 ( pid: IDENT DOUBLECOLON 
	 		{ pn = pn +(pn.length() == 0 ? "":"::")+pid.getText();}
	 )* 
	 {
	 	if (pn.length() > 0) {
		 	cr.setPackageName(pn);
		 	referencedPackages.add(pn);
	 	}
	 }
	 tid:IDENT {cr.setClassifierName(tid.getText()); cr.setFilename(this.getFilename());cr.setLine(tid.getLine());}
	( DOT iid:IDENT 
			{ cr.setClassifierName(tid.getText()+"."+iid.getText());}
	)?
    ;


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 9.3 Component Parameters
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

parameter returns [ Parameter p = FF.createParameter()]
       {Properties ps; ClassifierReference cr; processComments(p);}:
     ( IN OUT { p.setDirection(PortDirection.INOUT_LITERAL); }
     | IN { p.setDirection(PortDirection.IN_LITERAL); }
     | OUT { p.setDirection(PortDirection.OUT_LITERAL); }
     ) 
     PARAMETER
	 ( cr = classifier_reference {p.setClassifierReference(cr);} )?
	 ( ps = curlyPropertyAssociations_no_modes { p.setProperties(ps);}  )?
	 SEMI
//-------------------------------------------------------------------
;

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 9.4 Subcomponent Acess
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

dataAccess [AccessDirection d] returns [DataAccess da=null]
{Properties ps; ClassifierReference cr; }:
     DATA ACCESS 
         { da = FF.createDataAccess();  da.setDirection(d);processComments(da);}
     ( cr = classifier_reference {da.setClassifierReference(cr);} )?
	 ( ps = curlyPropertyAssociations_no_modes  { da.setProperties(ps);} )?
	 SEMI
//-------------------------------------------------------------------
;


busAccess [AccessDirection d] returns [BusAccess ba=null]
{Properties ps;ClassifierReference cr; }:
     BUS ACCESS 
         { ba = FF.createBusAccess();  ba.setDirection(d);processComments(ba);}
     ( cr = classifier_reference {ba.setClassifierReference(cr);} )?
	 ( ps = curlyPropertyAssociations_no_modes { ba.setProperties(ps);} )?
	 SEMI
//-------------------------------------------------------------------
;
subcomponent_access returns [Feature fe=null]:
    REQUIRES fe = dataAccess[AccessDirection.REQUIRED_LITERAL] 
    | REQUIRES fe = busAccess[AccessDirection.REQUIRED_LITERAL] 
    | PROVIDES fe = dataAccess[AccessDirection.PROVIDED_LITERAL] 
    | PROVIDES fe = busAccess[AccessDirection.PROVIDED_LITERAL] 
    ;
//-------------------------------------------------------------------
 



//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 9.5 Port Groups
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

port_group_type returns [PortGroupType pg=FF.createPortGroupType()]
   { PortGroupFeatures pgf=null;AnnexSubclause ax; Properties ps; ClassifierReference pgr, epg;
   boolean refined = false; Feature fs =null;RefinedReference fr;}:
    PORT GROUP defid:IDENT { pg.setName(defid.getText());
    		pg.setLocationReference(this.getFilename(),defid.getLine());
    		processComments(pg);}
    ( EXTENDS epg = unique_type_name 
            {pg.setExtendedClassifierReference(epg);}
    )?
	(( 
	  FEATURES
	  		{  pgf = FF.createPortGroupFeatures();
	  			pg.setFeatures(pgf);
	  		}
	    ( 
             id: IDENT 
             COLON 
             ( REFINED TO { refined = true; } )?
	         (fs = port_spec | fs = port_group_spec)
                { 
                	fs.setName(id.getText()); 
                	fs.setLocationReference(new LocationReference(this.getFilename(),id.getLine()));
                	fs.setRefined(refined); 
                    if (refined){
                    		 fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
                    		fs.setRefinedReference(fr);
                    		fr.setFilename(this.getFilename());fr.setLine(id.getLine());
                    		refined = false; }
                	pgf.addPortGroupElement(fs);
                }
	     )*
	  ( INVERSE OF pgr= unique_type_name { pg.setInverseClassifierReference(pgr);}
	   )?
	)
	|
	( INVERSE OF pgr= unique_type_name { pg.setInverseClassifierReference(pgr);} 
	))
	( ps = propertyAssociations_no_modes { pg.setProperties(ps); } )?
	( ax = annex_subclause { pg.addAnnexSubclause(ax); } )?
	END id2:IDENT SEMI
  { if (!( defid.getText().equalsIgnoreCase(id2.getText()) )) throw new SemanticException("Port group type name "+defid.getText()+" does not match "+id2.getText(), getFilename(),id2.getLine(), id2.getColumn());}
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(END);
			consumeUntil(SEMI);
			consume();
        }

port_group_spec returns [PortGroup pg=FF.createPortGroup()]
        {ClassifierReference pgt; Properties ps; processComments(pg);}:
    PORT GROUP
    ( pgt = unique_type_name { pg.setClassifierReference(pgt);} )?
    ( ps = curlyPropertyAssociations_no_modes { pg.setProperties(ps);} )?
    SEMI
//-------------------------------------------------------------------
;

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@ Section 10 CONNECTIONS
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


connectionsSubclause [ComponentImpl ci]
{Connections cns =null;}:		
     CONNECTIONS  
       {  cns = CnF.createConnections(); ci.setConnections(cns);processComments(cns);}
			(( connection[cns] | connection_refinement[cns] )+ | ( none_stmt ));

connection [Connections cns] 
{Properties ps=null; Connection cn=null; String cname=null;}:
     ( id:IDENT COLON  { cname = id.getText();})?
     ( cn = data_connection { cns.addDataConnection((DataConnection)cn);}
	 | cn = event_data_connection { cns.addEventDataConnection((EventDataConnection)cn);}
	 | cn = event_connection { cns.addEventConnection((EventConnection)cn);}
	 | cn = port_group_connection { cns.addPortGroupConnection((PortGroupConnection)cn);}
	 | cn = parameter_connection { cns.addParameterConnection((ParameterConnection)cn);}
	 | cn = bus_access_connection { cns.addBusAccessConnection((BusAccessConnection)cn);}
	 | cn = data_access_connection { cns.addDataAccessConnection((DataAccessConnection)cn);}
	 )
	     { processComments(cn);}
	    { if (cname != null) cn.setName(cname); }
	 ( ps = curlyPropertyAssociations { cn.setProperties(ps);} )?
 	 ( in_modes_and_transitions[cn] )? SEMI
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


data_connection returns [DataConnection dc=CnF.createDataConnection()]
        { FeatureReference cr; }:
	 lbl: DATA PORT cr = unique_port_identifier
	 	{ dc.setLocationReference(this.getFilename(),lbl.getLine());}
	    { dc.setSrcPortReference(cr);}
	 ( ARROW { dc.setTiming(ConnectionTiming.IMMEDIATE_LITERAL);}
	 | DARROW { dc.setTiming(ConnectionTiming.DELAYED_LITERAL);}
	 ) cr = unique_port_identifier
	    { dc.setDstPortReference(cr);}
//-------------------------------------------------------------------
;
event_connection returns [EventConnection ec=CnF.createEventConnection()]
        { FeatureReference cr; }:
	 lbl: EVENT PORT cr = unique_port_identifier
	 	{ ec.setLocationReference(this.getFilename(),lbl.getLine());}
	     { ec.setSrcPortReference(cr);}
	 ARROW 	 cr = unique_port_identifier
	     { ec.setDstPortReference(cr);}
//-------------------------------------------------------------------
;
event_data_connection returns [EventDataConnection edc=CnF.createEventDataConnection()]
        { FeatureReference cr; }:
	 lbl: EVENT DATA PORT cr = unique_port_identifier
	 	{ edc.setLocationReference(this.getFilename(),lbl.getLine());}
	     { edc.setSrcPortReference(cr);}
	 ARROW 	 cr = unique_port_identifier
        { edc.setDstPortReference(cr);}
//-------------------------------------------------------------------
;

parameter_connection returns [ParameterConnection pc=CnF.createParameterConnection()]
        { FeatureReference cr; }:
	 lbl: PARAMETER cr = unique_port_identifier
	 	{ pc.setLocationReference(this.getFilename(),lbl.getLine());}
	     { pc.setSrcPortReference(cr);}
	 ARROW 	 cr = unique_port_identifier
        { pc.setDstPortReference(cr);}
//-------------------------------------------------------------------
;

bus_access_connection returns [BusAccessConnection pc=CnF.createBusAccessConnection()]
        { FeatureReference cr; }:
	 lbl: BUS ACCESS cr = unique_port_identifier
	 	{ pc.setLocationReference(this.getFilename(),lbl.getLine());}
	     { pc.setSrcPortReference(cr);}
	 ARROW 	 cr = unique_port_identifier
        { pc.setDstPortReference(cr);}
//-------------------------------------------------------------------
;


data_access_connection returns [DataAccessConnection pc=CnF.createDataAccessConnection()]
        { FeatureReference cr; }:
	 lbl: DATA ACCESS cr = unique_port_identifier
	 	{ pc.setLocationReference(this.getFilename(),lbl.getLine());}
	     { pc.setSrcPortReference(cr);}
	 ARROW 	 cr = unique_port_identifier
        { pc.setDstPortReference(cr);}
//-------------------------------------------------------------------
;


port_group_connection returns [PortGroupConnection sc = CnF.createPortGroupConnection()]
        { FeatureReference cr; }:
	 lbl: PORT GROUP
	 	{ sc.setLocationReference(this.getFilename(),lbl.getLine());}
	 cr = unique_port_group_identifier
	     { sc.setSrcPortReference(cr);}
	 ARROW  cr = unique_port_group_identifier
	     { sc.setDstPortReference(cr);}
;

unique_port_group_identifier returns [FeatureReference cr = new FeatureReference()]:
 	compid:IDENT {cr.setCompName(compid.getText());}
 	DOT portid:IDENT {cr.setFeatureName(portid.getText());cr.setFilename(this.getFilename());cr.setLine(portid.getLine());}
   |
   soleportid:IDENT {cr.setFeatureName(soleportid.getText());cr.setCompName("");cr.setFilename(this.getFilename());cr.setLine(soleportid.getLine());}
//	 subcomponent_identifier:IDENT DOT port_group_identifier:IDENT
//	 |
//	 component_type_port_group_identifier:IDENT DOT element_port_group_identifier:IDENT
//	 |
//	 component_type_port_group_identifier:IDENT  // reordered because of no dot
;

connection_refinement[Connections cns] 
{Properties ps; Connection cn = null;RefinedReference fr;}:
     id:IDENT COLON REFINED TO
     ( EVENT DATA PORT { cn = CnF.createEventDataConnection();cns.addEventDataConnection((EventDataConnection)cn);}
     | EVENT PORT { cn = CnF.createEventConnection();cns.addEventConnection((EventConnection)cn);}
     | DATA PORT { cn = CnF.createDataConnection();cns.addDataConnection((DataConnection)cn);}
     | PORT GROUP { cn = CnF.createPortGroupConnection();cns.addPortGroupConnection((PortGroupConnection)cn);}
     | PARAMETER { cn = CnF.createParameterConnection();cns.addParameterConnection((ParameterConnection)cn);}
     | BUS ACCESS { cn = CnF.createBusAccessConnection();cns.addBusAccessConnection((BusAccessConnection)cn);}
     | DATA ACCESS { cn = CnF.createDataAccessConnection();cns.addDataAccessConnection((DataAccessConnection)cn);}
     )
     	{ cn.setName(id.getText()); }
         { cn.setRefined(true); processComments(cn);}
                {  fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
                	cn.setRefinedReference(fr);
               		fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
          		}
	 (( ps = curlyPropertyAssociations 
	      { cn.setProperties(ps); }
	 	( in_modes_and_transitions[cn] )?
	  ) |
	  in_modes_and_transitions[cn]
	 )
	 SEMI
;

dataConnectionsSubclause [ComponentImpl ci]
{Connections cns =null; Connection cn = null;}:		
     CONNECTIONS  
       {  cns = CnF.createConnections(); ci.setConnections(cns);processComments(cns);}
			(( cn = data_access_connection_decl { cns.addDataAccessConnection((DataAccessConnection)cn);}
	         | cn = data_access_connection_refinement_decl { cns.addDataAccessConnection((DataAccessConnection)cn);}
	         )+ | ( none_stmt ));

data_access_connection_decl returns [Connection cn = null] 
{Properties ps=null; String cname=null;}:
     ( id:IDENT COLON  { cname = id.getText();})?
	 cn = data_access_connection 
	     { processComments(cn);}
	    { if (cname != null) cn.setName(cname); }
	 ( ps = curlyPropertyAssociations { cn.setProperties(ps);} )?
 	 ( in_modes[cn] )? SEMI
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


data_access_connection_refinement_decl returns [Connection cn = null] 
{Properties ps; RefinedReference fr;}:
     id:IDENT COLON REFINED TO
      DATA ACCESS { cn = CnF.createDataAccessConnection();}
     	{ cn.setName(id.getText()); }
         { cn.setRefined(true); processComments(cn);}
                {  fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
                	cn.setRefinedReference(fr);
               		fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
          		}
	 (( ps = curlyPropertyAssociations 
	      { cn.setProperties(ps); }
	 	( in_modes[cn] )?
	  ) |
	  in_modes[cn]
	 )
	 SEMI
;


busConnectionsSubclause [ComponentImpl ci]
{Connections cns =null; Connection cn = null;}:		
     CONNECTIONS  
       {  cns = CnF.createConnections(); ci.setConnections(cns);processComments(cns);}
			(( cn = bus_access_connection_decl { cns.addBusAccessConnection((BusAccessConnection)cn);}
	         | cn = bus_access_connection_refinement_decl { cns.addBusAccessConnection((BusAccessConnection)cn);}
	         )+ | ( none_stmt ));


bus_access_connection_decl returns [Connection cn = null] 
{Properties ps=null; String cname=null;}:
     ( id:IDENT COLON  { cname = id.getText();})?
	 cn = bus_access_connection 
	     { processComments(cn);}
	    { if (cname != null) cn.setName(cname); }
	 ( ps = curlyPropertyAssociations { cn.setProperties(ps);} )?
 	 ( in_modes[cn] )? SEMI
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

bus_access_connection_refinement_decl returns [Connection cn = null] 
{Properties ps; RefinedReference fr;}:
     id:IDENT COLON REFINED TO
      BUS ACCESS { cn = CnF.createBusAccessConnection();}
     	{ cn.setName(id.getText()); }
         { cn.setRefined(true); processComments(cn);}
                {  fr = new RefinedReference(); fr.setRefinedName(id.getText()); 
                	cn.setRefinedReference(fr);
               		fr.setFilename(this.getFilename());fr.setLine(id.getLine()); 
          		}
	 (( ps = curlyPropertyAssociations 
	      { cn.setProperties(ps); }
	 	( in_modes[cn] )?
	  ) |
	  in_modes[cn]
	 )
	 SEMI
;


//-------------------------------------------------------------------
//@
//@   End-To-End Flows
//@

flow_specs[ComponentType ct]
{FlowSpecs flows=null;}: 
    FLOWS {  flows = FlF.createFlowSpecs(); ct.setFlowSpecs(flows);processComments(flows); }
      ( (flow_spec[flows])+
      | none_stmt
      )
;

flow_spec [FlowSpecs flows]:
     defining_identifier:IDENT 
     COLON 
     ( flow_source_spec[flows,defining_identifier]
     | flow_sink_spec[flows,defining_identifier]
     | flow_path_spec[flows,defining_identifier] 
     |  ( REFINED TO 
         ( flow_source_spec_refinement[flows,defining_identifier]
         | flow_sink_spec_refinement[flows,defining_identifier]
         | flow_path_spec_refinement[flows,defining_identifier] 
         )
       )
     )
	 SEMI
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }


//-------------------------------------------------------------------
flow_source_spec [FlowSpecs fss, Token defining_identifier]
{Properties ps; FeatureReference cr;FlowSourceSpec flow=null;}: 
	FLOW SOURCE 
        { flow = FlF.createFlowSourceSpec();}
        { fss.addFlowSourceSpec(flow); processComments(flow);}
		{ flow.setName(defining_identifier.getText()); flow.setLocationReference(new LocationReference(this.getFilename(),defining_identifier.getLine())); }
     cr = flow_feature_identifier
	     { flow.setDstPortReference(cr);}
	 ( ps = curlyPropertyAssociations_no_modes 
	     { flow.setProperties(ps); }
	     )?
;

flow_source_spec_refinement [FlowSpecs fss, Token defining_identifier]
{Properties ps; FeatureReference cr;FlowSourceSpec flow=null;RefinedReference fr;}: 
	FLOW SOURCE 
        { flow = FlF.createFlowSourceSpec();}
        { fss.addFlowSourceSpec(flow); processComments(flow);}
		{ flow.setName(defining_identifier.getText()); flow.setLocationReference(new LocationReference(this.getFilename(),defining_identifier.getLine()));
           flow.setRefined(true);
           fr = new RefinedReference(); 
           fr.setRefinedName(defining_identifier.getText()); 
           fr.setFilename(this.getFilename());fr.setLine(defining_identifier.getLine()); 
           flow.setRefinedReference(fr);}
	  ps = curlyPropertyAssociations_no_modes 
	     { flow.setProperties(ps); }
	     
;


flow_sink_spec [FlowSpecs fss, Token defining_identifier]
{Properties ps; FeatureReference cr; FlowSinkSpec flow=null;}: 
     FLOW SINK 
        {flow= FlF.createFlowSinkSpec();
        fss.addFlowSinkSpec(flow); processComments(flow);
        flow.setName(defining_identifier.getText()); flow.setLocationReference(new LocationReference(this.getFilename(),defining_identifier.getLine()));}

     cr = flow_feature_identifier
	     { flow.setSrcPortReference(cr);}
	 ( ps = curlyPropertyAssociations_no_modes 
	     { flow.setProperties(ps); }
	 )?
;


flow_sink_spec_refinement [FlowSpecs fss, Token defining_identifier]
{Properties ps; FeatureReference cr; FlowSinkSpec flow=null;RefinedReference fr;}: 
     FLOW SINK 
        { flow= FlF.createFlowSinkSpec();
        fss.addFlowSinkSpec(flow); processComments(flow);
        flow.setName(defining_identifier.getText()); flow.setLocationReference(new LocationReference(this.getFilename(),defining_identifier.getLine()));
        flow.setRefined(true);
           fr = new RefinedReference(); 
           fr.setRefinedName(defining_identifier.getText()); 
           fr.setFilename(this.getFilename());fr.setLine(defining_identifier.getLine()); 
           flow.setRefinedReference(fr); }
	  ps = curlyPropertyAssociations_no_modes 
	     { flow.setProperties(ps); }
	 
;

flow_path_spec [FlowSpecs fss, Token defining_identifier]
{Properties ps; FeatureReference cr; FlowPathSpec flow=null;}: 
	FLOW PATH
        { flow= FlF.createFlowPathSpec();
        fss.addFlowPathSpec(flow); processComments(flow);
        flow.setName(defining_identifier.getText()); flow.setLocationReference(new LocationReference(this.getFilename(),defining_identifier.getLine()));}

      cr = flow_feature_identifier ARROW 
	     { flow.setSrcPortReference(cr);}
      cr = flow_feature_identifier
	     { flow.setDstPortReference(cr);}
	 ( ps = curlyPropertyAssociations_no_modes 
	     { flow.setProperties(ps); }
	 )?
;

flow_path_spec_refinement [FlowSpecs fss, Token defining_identifier]
{Properties ps; FeatureReference cr; FlowPathSpec flow=null;RefinedReference fr;}: 
	FLOW PATH
        { flow= FlF.createFlowPathSpec();
        fss.addFlowPathSpec(flow); processComments(flow);
        flow.setName(defining_identifier.getText()); flow.setLocationReference(new LocationReference(this.getFilename(),defining_identifier.getLine()));
        flow.setRefined(true);
           fr = new RefinedReference(); 
           fr.setRefinedName(defining_identifier.getText()); 
           fr.setFilename(this.getFilename());fr.setLine(defining_identifier.getLine()); 
           flow.setRefinedReference(fr);  }
	  ps = curlyPropertyAssociations_no_modes 
	     { flow.setProperties(ps); }
;


flow_impls[ComponentImpl ci]
{ Flows flows=null;}: 
    FLOWS {  flows = FlF.createFlows(); ci.setFlows(flows); processComments(flows);}
      ( ( flow_sequence[flows]  )+ 
      | none_stmt
      )
;

flow_sequence [Flows flows] 
:
     defining_identifier:IDENT 
     COLON 
     ( flow_source_implementation[flows,defining_identifier]
     | flow_sink_implementation[flows,defining_identifier]
     | flow_path_implementation[flows,defining_identifier] 
     | end_to_end_flow[flows,defining_identifier] 
     |  ( REFINED TO 
         ( flow_source_implementation_refinement[flows,defining_identifier]
         | flow_sink_implementation_refinement[flows,defining_identifier]
         | flow_path_implementation_refinement[flows,defining_identifier] 
         | end_to_end_flow_refinement[flows,defining_identifier] 
         )
       )
     )
	 SEMI
;
		exception
          catch [RecognitionException ex] {
            reportError(ex);
			consumeUntil(SEMI);
			consume();
        }

flow_source_implementation [Flows flows, Token flow_identifier]
{Properties ps; FlowSourceImpl fi=null;FeatureReference cr; LocationReference fr;FlowElement fe=null;}:
     FLOW SOURCE
        { fi = FlF.createFlowSourceImpl();}
		{ flows.addFlowSourceImpl(fi); processComments(fi);}
         { fi.setName(flow_identifier.getText());}
         {  fr = new LocationReference(this.getFilename(),flow_identifier.getLine());fi.setLocationReference(fr);}

       ( cr = subcomponent_flow_identifier 
         {  
         	fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr);
         }
       ARROW cr = connection_identifier 
       		{ 
       			fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr);}
       ARROW )*
       cr = flow_feature_identifier 
       { fi.setDstPortReference(cr); }
	 ( ps = curlyPropertyAssociations 
	     { fi.setProperties(ps);}
	   ( in_modes[fi] )?
	   |  
       in_modes[fi] )?
;

flow_sink_implementation [Flows flows, Token flow_identifier]
{Properties ps; FlowSinkImpl fi=null;FeatureReference cr; LocationReference fr;FlowElement fe=null;}:
      FLOW SINK
        { fi = FlF.createFlowSinkImpl();}
		{ flows.addFlowSinkImpl(fi);processComments(fi);}
         { fi.setName(flow_identifier.getText());}
         {  fr = new LocationReference(this.getFilename(),flow_identifier.getLine());fi.setLocationReference(fr);}
      cr = flow_feature_identifier 
     { fi.setSrcPortReference(cr); }
     ( ARROW cr = connection_identifier 
         {  fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr);}
     ARROW cr = subcomponent_flow_identifier { fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr);}
     )*
	 ( ps = curlyPropertyAssociations 
	     { fi.setProperties(ps);}
	   ( in_modes[fi] )?
	   |  
       in_modes[fi] )?
;

flow_path_implementation [Flows flows, Token flow_identifier]
{Properties ps; FlowPathImpl fi=null;FlowElement fe=null;FeatureReference cr; LocationReference fr;}:
       FLOW PATH 
        { fi = FlF.createFlowPathImpl();}
		{ flows.addFlowPathImpl(fi);processComments(fi);}
        { fi.setName(flow_identifier.getText());}
         {  fr = new LocationReference(); fr.setFilename(this.getFilename()); 
         	fr.setLine(flow_identifier.getLine());fi.setLocationReference(fr);}
       cr = flow_feature_identifier 
       { fi.setSrcPortReference(cr); }
       ARROW 
       ( cr = connection_identifier 
             {  fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr);}
         ARROW
         (  cr = subcomponent_flow_identifier 
             {   fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr);}
         ARROW cr = connection_identifier 
         	{ fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr);} 
         ARROW )+
       )?
      cr = flow_feature_identifier 
      { fi.setDstPortReference(cr); }

	 ( ps = curlyPropertyAssociations 
	     { fi.setProperties(ps);}
	   ( in_modes[fi] )?
	   |  
       in_modes[fi] )?
;

flow_source_implementation_refinement [Flows flows, Token flow_identifier]
        	{Properties ps; FlowSourceImpl fi=null;LocationReference lr;RefinedReference fr;}:
     FLOW SOURCE
		{ fi = FlF.createFlowSourceImpl();}
		{ flows.addFlowSourceImpl(fi); }
			{ processComments(fi);}
         { fi.setName(flow_identifier.getText());}
         {  lr = new LocationReference(); lr.setFilename(this.getFilename()); 
         	lr.setLine(flow_identifier.getLine());fi.setLocationReference(lr);}
         { fi.setRefined(true);
            fr = new RefinedReference(); 
           fr.setRefinedName(flow_identifier.getText()); 
           fr.setFilename(this.getFilename());fr.setLine(flow_identifier.getLine()); 
           fi.setRefinedReference(fr); 
         }

	 ( ps = curlyPropertyAssociations 
	     { fi.setProperties(ps);}
	   ( in_modes[fi] )?
	   |  
       in_modes[fi] )?
;

flow_sink_implementation_refinement [Flows flows, Token flow_identifier]
        	{Properties ps; FlowSinkImpl fi=null;LocationReference lr;RefinedReference fr;}:
     FLOW SINK
		{ fi = FlF.createFlowSinkImpl();
		 flows.addFlowSinkImpl(fi);
			processComments(fi);
         fi.setName(flow_identifier.getText());
          lr = new LocationReference(); lr.setFilename(this.getFilename()); 
         lr.setLine(flow_identifier.getLine());fi.setLocationReference(lr);
          fi.setRefined(true);
            fr = new RefinedReference(); 
           fr.setRefinedName(flow_identifier.getText()); 
           fr.setFilename(this.getFilename());fr.setLine(flow_identifier.getLine()); 
           fi.setRefinedReference(fr); 
         }

	 ( ps = curlyPropertyAssociations 
	     { fi.setProperties(ps);}
	   ( in_modes[fi] )?
	   |  
       in_modes[fi] )?
;

flow_path_implementation_refinement [Flows flows, Token flow_identifier]
        	{Properties ps; FlowPathImpl fi=null;LocationReference lr;RefinedReference fr;}:
     FLOW PATH
		{ fi = FlF.createFlowPathImpl();}
			{ processComments(fi);}
		{ flows.addFlowPathImpl(fi); }
        { fi.setName(flow_identifier.getText());}
         {  lr = new LocationReference(); lr.setFilename(this.getFilename()); 
         	lr.setLine(flow_identifier.getLine());fi.setLocationReference(lr);}
         { fi.setRefined(true);
            fr = new RefinedReference(); 
           fr.setRefinedName(flow_identifier.getText()); 
           fr.setFilename(this.getFilename());fr.setLine(flow_identifier.getLine()); 
           fi.setRefinedReference(fr); 
         }

	 ( ps = curlyPropertyAssociations 
	     { fi.setProperties(ps);}
	   ( in_modes[fi] )?
	   |  
       in_modes[fi] )?
;

subcomponent_flow_identifier  returns [FeatureReference cr = new FeatureReference()] :
 	subcompid:IDENT {cr.setCompName(subcompid.getText());}
 	DOT flowid:IDENT {cr.setFeatureName(flowid.getText());cr.setFilename(this.getFilename());cr.setLine(flowid.getLine());}
;

end_to_end_flow [Flows flows, Token flow_identifier]
        {Properties ps; EndToEndFlow fi=null;FeatureReference cr; LocationReference lr; FlowElement fe=null;}:
     END TO END FLOW  
        { fi = FlF.createEndToEndFlow(); flows.addEndToEndFlow(fi);}
        { processComments(fi);}
        { fi.setName(flow_identifier.getText());}
         {  lr = new LocationReference(); lr.setFilename(this.getFilename()); 
         	lr.setLine(flow_identifier.getLine());fi.setLocationReference(lr);}
     cr = subcomponent_flow_identifier 
     		{ fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr);
     		}
     ( ARROW cr = connection_identifier 
         {  fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr); }
     ARROW cr = subcomponent_flow_identifier { fe = FlF.createFlowElement(); fi.addFlowElement(fe);
     			processComments(fe);
     			fe.setFeatureReference(cr);  }
     )+
	 ( ps = curlyPropertyAssociations { fi.setProperties(ps);}
	 )? 
    ( in_modes[fi] )?
    { processComments(fi); }
;

end_to_end_flow_refinement [Flows flows, Token flow_identifier]
        	{Properties ps; EndToEndFlow fi=null;LocationReference lr;RefinedReference fr;}:
     END TO END FLOW 
		{ fi = FlF.createEndToEndFlow();}
			{ processComments(fi);}
		{ flows.addEndToEndFlow(fi); }
         { fi.setName(flow_identifier.getText());}
         {  lr = new LocationReference(); lr.setFilename(this.getFilename()); 
         	lr.setLine(flow_identifier.getLine());fi.setLocationReference(lr);}
         { fi.setRefined(true);
            fr = new RefinedReference(); 
           fr.setRefinedName(flow_identifier.getText()); 
           fr.setFilename(this.getFilename());fr.setLine(flow_identifier.getLine()); 
           fi.setRefinedReference(fr); 
         }

	 ( ps = curlyPropertyAssociations 
	     { fi.setProperties(ps);}
	   ( in_modes[fi] )?
	   |  
       in_modes[fi] )?
;

connection_identifier returns [FeatureReference cr = new FeatureReference()] :
   connid:IDENT {cr.setFeatureName(connid.getText());cr.setCompName("");cr.setFilename(this.getFilename());cr.setLine(connid.getLine());}
;

flow_feature_identifier returns [FeatureReference cr = new FeatureReference()] :
 	compid:IDENT {cr.setCompName(compid.getText());}
 	DOT portid:IDENT {cr.setFeatureName(portid.getText());cr.setFilename(this.getFilename());cr.setLine(portid.getLine());}
   |
   soleportid:IDENT {cr.setFeatureName(soleportid.getText());cr.setCompName("");cr.setFilename(this.getFilename());cr.setLine(soleportid.getLine());}
//     port_identifier
//     | parameter_identifier
//     | port_group_identifier
//     | port_group_identifier DOT port_identifier
//-----------------------------------------------------
;

real_literal returns [Token tok=null]
{String str;}:
    id:NUMERIC_LIT
	{ 
		str = id.getText();
		tok = id;
        if ( str.indexOf('.') == -1 ) throw new SemanticException("Expected a real value: "+str,getFilename(),id.getLine(), id.getColumn()); }
	;
	
integer_literal returns [Token tok =null] 
{String str;}:
    id:NUMERIC_LIT
	{ 
		 str = id.getText();
		tok = id;
         if ( str.indexOf('.') != -1 ) throw new SemanticException("Expected an integer value: "+str,getFilename(),id.getLine(), id.getColumn()); }
	;

{	
	import edu.cmu.sei.aadl.model.pluginsupport.ParseErrorReporter;
	import org.eclipse.core.resources.IResource; 
}
class AadlLexer extends Lexer;

options {
    k=3; // needed for newline junk
    charVocabulary='\u0000'..'\uFFFE'; // allow UNICODE
    caseSensitive = false;
    caseSensitiveLiterals = false;
	defaultErrorHandler=true;
}

tokens {
	AADLSPEC;
	ACCESS="access";
	AND="and";
	ALL="all";
	ANNEX="annex";
	APPLIES="applies";
	BINDING="binding";
	BOOLEAN="aadlboolean";
	BUS="bus";
	CALLS="calls";
	CLASSIFIER="classifier";
	REFERENCE="reference";
	CONNECTIONS="connections";
	CONSTANT="constant";
	DATA="data";
	DELTA="delta";
	DEVICE="device";
	END="end";
	ENUMERATION="enumeration";
	EVENT="event";
	EXTENDS="extends";
	FALSE="false";
	FEATURES="features";
	FLOW="flow";
	FLOWS="flows";
	GROUP="group";
	IMPL="implementation";
	IN="in";
	INHERIT="inherit";
	INITIAL="initial";
	INTEGER="aadlinteger";
	INVERSE="inverse";
	IS="is";
	LIST="list";
	MEMORY="memory";
	MODE="mode";
	MODES="modes";
	NONE = "none";
	NOT="not";
	OF="of";
	OR="or";
	OUT="out";
	PACKAGE="package";
	PARAMETER="parameter";
	PATH="path";
	PORT="port";
	PRIVATE="private";
	PROCESS="process";
	PROCESSOR="processor";
	PROPERTIES="properties";
	PROPERTY="property";
	PROVIDES="provides";
	PUBLIC="public";
	RANGE="range";
	REAL="aadlreal";
	REFINED="refined";
	REFINES="refines";
	REQUIRES = "requires";
	SERVER="server";
	SET = "set";
	SINK="sink";
	SOURCE="source";
	STRING="aadlstring";
	SUBCOMPONENTS="subcomponents";
	SUBPROGRAM="subprogram";
	SYSTEM="system";
	THREAD="thread";
	TO="to";
	TRANSITIONS="transitions";
	TRUE="true";
	TYPE="type";
	UNITS="units";
	VALUE="value";
	
}

{
	
	/**
	 * the current resource being parsed
	 */
	protected ParseErrorReporter errReporter = null;
	
	/**
	 * Set the Parse Error Reporter with which to produce error messages.
	 */
	public void setParseErrorReporter(final ParseErrorReporter reporter) {
		  errReporter = reporter;
	}
		
	/**
	 * Report lexer error
	 * @param e  antlr error exception 
	 */
	public  void reportError(RecognitionException e) {
		  errReporter.error(e.getFilename(), e.getLine(), e.getMessage());
	}
	
	/**
	 * Dummy method
	 * see http://www.antlr.org:8080/pipermail/antlr-interest/2005-August/013323.html and replies
	 */
	 void recover (RecognitionException e, BitSet b) {
	 }
}
 
LPAREN: '(' ;
RPAREN: ')' ;
LCURLY: '{' ;
RCURLY: '}' ;
COLON:  ':' ;
PLUS  : '+' ;
MINUS : '-' ;
STAR  : '*' ;
SEMI  : ';' ;
COMMA : ',' ;
DOT	  : '.' ;
DOTDOT : ".." ;
ASSIGN : "=>";
ASSIGNPLUS : "+=>";
DOUBLECOLON: "::";
LTRANS : "-[";
RTRANS : "]->";
ARROW : "->" ;
DARROW: "->>";
HASH: '#';
  
IDENT:	('a'..'z'
        //|'A'..'Z'
        ) ( ('_')? ('a'..'z'
        //|'A'..'Z'
        |'0'..'9'))*
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consume();
          }
	;

// string literals
STRING_LITERAL
	:	'"' ("\"\"" |ESC|~('"'|'\\'))* '"'
	;


NUMERIC_LIT : 
		( DIGIT )+
		(( '#' BASED_INTEGER  '#' ( INT_EXPONENT )? )
		| ( '_' ( DIGIT )+ )*  // INTEGER
		  ( { LA(2)!='.' }?  //&& LA(3)!='.' }?
				// real
			( '.' ( DIGIT )+ ( '_' ( DIGIT )+ )* ( EXPONENT )?)
				// integer with exponent
			| (INT_EXPONENT)?
		  )
		)
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consume();
          }
  ;



// a couple protected methods to assist in matching the various numbers

protected
DIGIT   :  ( '0'..'9' ) ;

protected
EXPONENT           :  ('e') ('+'|'-')? ( DIGIT )+ ;

protected
INT_EXPONENT           :  ('e') ('+')? ( DIGIT )+ 
			exception
          catch [RecognitionException ex] {
            reportError(ex);
			consume();
          }
;


protected
EXTENDED_DIGIT     :  ( DIGIT | 'a'..'f' ) ;

protected
BASED_INTEGER      :  ( EXTENDED_DIGIT ) ( ('_')? EXTENDED_DIGIT )* ;

protected
BASE				: DIGIT ( DIGIT )?
;

// escape sequence -- note that this is protected; it can only be called
//   from another lexer rule -- it will not ever directly return a token to
//   the parser
// There are various ambiguities hushed in this rule.  The optional
// '0'...'9' digit matches should be matched here rather than letting
// them go back to STRING_LITERAL to be matched.  ANTLR does the
// right thing by matching immediately; hence, it's ok to shut off
// the FOLLOW ambig warnings.
protected
ESC
	:	'\\'
		(	'n'
		|	'r'
		|	't'
		|	'b'
		|	'f'
		|	'"'
		|	'\''
		|	'\\'
		|	('u')+ HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT
		|	'0'..'3'
			(
				options {
					warnWhenFollowAmbig = false;
				}
			:	'0'..'7'
				(
					options {
						warnWhenFollowAmbig = false;
					}
				:	'0'..'7'
				)?
			)?
		|	'4'..'7'
			(
				options {
					warnWhenFollowAmbig = false;
				}
			:	'0'..'7'
			)?
		)
	;


// hexadecimal digit (again, note it's protected)
protected
HEX_DIGIT
	:	('0'..'9'|
	//'A'..'F'|
	'a'..'f')
	;


WS    : ( ' '
        | '\r' '\n' {newline();}
        | '\r' {newline();}
        | '\n' {newline();}
        | '\t' {tab();}
        )
        {$setType(Token.SKIP);}
      ;    
    

// multiple-line comments
// if we change the comment characters, we need to fix the unparser
// Unparser supports conversion of multi-line into single line comments.
//ML_COMMENT
//	:	"/*"
//		(	/*	'\r' '\n' can be matched in one alternative or by matching
//				'\r' in one iteration and '\n' in another.  I am trying to
//				handle any flavor of newline that comes in, but the language
//				that allows both "\r\n" and "\r" and "\n" to all be valid
//				newline is ambiguous.  Consequently, the resulting grammar
//				must be ambiguous.  I'm shutting this warning off.
//			 */
//			options {
//				generateAmbigWarnings=false;
//			}
//		:
//			{ LA(2)!='/' }? '*'
//			//{ LA(2)!='*' || LA(3) != '}' }? '*'
//		|	'\r' '\n'		{newline();}
//		|	'\r'			{newline();}
//		|	'\n'			{newline();}
//		|	~('*'|'\n'|'\r')
//		)*
//		"*/"
//	;
     

// Annex text processing: derived from multiple-line comments
ANNEX_TEXT
	:	"{**"
		(	/*	'\r' '\n' can be matched in one alternative or by matching
				'\r' in one iteration and '\n' in another.  I am trying to
				handle any flavor of newline that comes in, but the language
				that allows both "\r\n" and "\r" and "\n" to all be valid
				newline is ambiguous.  Consequently, the resulting grammar
				must be ambiguous.  I'm shutting this warning off.
			 */
			options {
				generateAmbigWarnings=false;
			}
		:
			{ LA(2)!='*' || LA(3) != '}' }? '*'
		|	'\r' '\n'		{newline();}
		|	'\r'			{newline();}
		|	'\n'			{newline();}
		|	~('*'|'\n'|'\r')
		)*
		"**}"
	;
 
     
// Single-line comments
SL_COMMENT
	:	"--"
(~('\n'|'\r' ))* 
	;
 
 